# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

ksuid_edge = {
    'color': {
        'color': 'orange'
    },
    'title': 'KSUID Parsing Functions',
    'label': 'KSUID'
}


EPOCH_OFFSET = 1400000000  # Offset from January 1, 1970, so that the lifespan of KSUIDs will be considerably longer
TIMESTAMP_LENGTH = 4  # 4 bytes are storing the timestamp = 8 characters
KSUID_LENGTH = 27  # length of KSUID after encoding
CHARSET = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"
BASE = 62


def run(unfurl, node):

    # From https://github.com/segmentio/ksuid:
    #   KSUID is for K-Sortable Unique IDentifier. It's a way to generate globally unique IDs
    #   similar to RFC 4122 UUIDs, but contain a time component so they can be "roughly" sorted
    #   by time of creation. The remainder of the KSUID is randomly generated bytes.
    #
    # Ref:
    #   - https://github.com/segmentio/ksuid
    #   - https://github.com/saresend/KSUID

    if node.data_type == 'ksuid':
        decoded_str_b62_to_bytes = decodebytes(node.value)
        time_part_in_bytes = decoded_str_b62_to_bytes[0:TIMESTAMP_LENGTH]
        timestamp = EPOCH_OFFSET + int.from_bytes(time_part_in_bytes, 'big', signed=False)
        random_payload = decoded_str_b62_to_bytes[TIMESTAMP_LENGTH:]

        unfurl.add_to_queue(
            data_type='epoch-seconds', key=None, value=timestamp, label=f'Timestamp: {timestamp}',
            parent_id=node.node_id, incoming_edge_config=ksuid_edge)

        unfurl.add_to_queue(
            data_type='descriptor', key=None, value=random_payload,
            label=f'Randomly generated payload: {random_payload}',
            parent_id=node.node_id, incoming_edge_config=ksuid_edge)

    else:
        m = re.match(r'([a-zA-Z0-9]{27})', str(node.value))
        if m and len(node.value) == 27:
            # If timestamp component between 2014 and 2027
            if '090000l1tmebfs0000000000000' < str(node.value) < '3WgEPTl1tmebfsQzFP4bxwgy80V':
                unfurl.add_to_queue(
                    data_type='ksuid', key=None, value=node.value, label=f'KSUID: {node.value}',
                    hover='KSUID are identifiers that are comprised of a timestamp and a random number. '
                          '<a href="https://github.com/segmentio/ksuid" target="_blank">[ref]</a>',
                    parent_id=node.node_id, incoming_edge_config=ksuid_edge,
                    extra_options={'widthConstraint': {'maximum': 300}})


def decode_b62(b):
    """Decodes a base62 encoded value ``b``."""

    if b.startswith("0z"):
        b = b[2:]

    l, i, v = len(b), 0, 0
    for x in b:
        v += _value(x) * (BASE ** (l - (i + 1)))
        i += 1

    return v


def decodebytes(s):
    """Decodes a string of base62 data into a bytes object.
    :param s: A string to be decoded in base62
    :rtype: bytes
    """

    decoded = decode_b62(s)
    buf = bytearray()
    while decoded > 0:
        buf.append(decoded & 0xff)
        decoded //= 256
    buf.reverse()

    return bytes(buf)


def _value(ch):
    """Decodes an individual digit of a base62 encoded string."""

    try:
        return CHARSET.index(ch)
    except ValueError:
        raise ValueError("base62: Invalid character (%s)" % ch)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

import logging
log = logging.getLogger(__name__)

sonyflake_edge = {
    'color': {
        'color': 'red'
    },
    'title': 'Sonyflake Parsing Functions',
    'label': '❄'
}


def parse_sonyflake(unfurl, node):
    # Ref: https://github.com/sony/sonyflake

    try:
        snowflake = int(node.value, 16)
        ts = (snowflake >> 24)
        timestamp = ts + 140952960000
        sequence = (snowflake >> 16) & 0xFF
        machine_id_1 = (snowflake >> 8) & 0xFF
        machine_id_2 = snowflake & 0xFF

    except Exception as e:
        log.exception(f'Exception parsing snowflake: {e}')
        return

    node.hover = 'Sonyflake is a distributed unique ID generator inspired by ' \
                 'Twitter\'s Snowflake. <a href="https://github.com/sony/sonyflake" ' \
                 'target="_blank">[ref]</a>'

    unfurl.add_to_queue(
        data_type='epoch-centiseconds', key=None, value=timestamp, label=f'Timestamp: {timestamp}',
        hover='The first value in a Sonyflake is a timestamp',
        parent_id=node.node_id, incoming_edge_config=sonyflake_edge)

    unfurl.add_to_queue(
        data_type='integer', key=None, value=sequence, label=f'Sequence: {sequence}',
        hover='For every ID that is generated, this number is incremented and rolls over every 256',
        parent_id=node.node_id, incoming_edge_config=sonyflake_edge)

    unfurl.add_to_queue(
        data_type='integer', key=None, value=machine_id_1, label=f'Machine ID: {machine_id_1}.{machine_id_2}',
        hover='The second value in a Sonyflake is the machine ID. By default this is '
              '<br>the <b>lower half of the private IP address</b> of the system generating the ID',
        parent_id=node.node_id, incoming_edge_config=sonyflake_edge)


def run(unfurl, node):
    if not node.data_type.startswith(('sonyflake', 'hash')):

        long_int = re.fullmatch(r'\d{15}', str(node.value))
        # Sonyflakes should be 15 hex digits long; limiting them to first digit 1-9 limits time frame from 2016 to 2026.
        m = re.fullmatch(r'(?P<sonyflake>[1-9][A-F0-9]{14})', str(node.value).replace('-', '').upper())
        if m and not long_int:
            parse_sonyflake(unfurl, node)

    elif node.data_type == 'sonyflake':
        parse_sonyflake(unfurl, node)



from .types import type_maps


class Config:
    def __init__(self):

        # Map of message type names to typedefs, previously stored at
        # `blackboxprotobuf.known_messages`
        self.known_types = {}

        # Default type for "bytes" like objects that aren't messages or strings
        # Other option is currently just 'bytes_hex'
        self.default_binary_type = "bytes"

        # Change the default type for a wiretype (eg.
        self.default_types = {}

    def get_default_type(self, wiretype):
        default_type = self.default_types.get(wiretype, None)
        if default_type is None:
            default_type = type_maps.WIRE_TYPE_DEFAULTS.get(wiretype, None)

        return default_type


default = Config()



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = "Ryan Benson"
__version__ = "2022.02"
__email__ = "ryan@dfir.blog"

import logging
import sys

log = logging.getLogger(__name__)
log.setLevel('WARNING')
handler = logging.StreamHandler(sys.stdout)
handler.setLevel(logging.DEBUG)
formatter = logging.Formatter(
    '%(asctime)s.%(msecs).03d | %(name)s | %(levelname)s | %(message)s', datefmt='%Y-%m-%d %H:%M:%S')
handler.setFormatter(formatter)
log.addHandler(handler)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
log = logging.getLogger(__name__)

discord_edge = {
    'color': {
        'color': '#7289da'
    },
    'title': 'Discord Snowflake',
    'label': '❄'
}


def parse_discord_snowflake(unfurl, node):
    try:
        snowflake = int(node.value)
    # Some non-integer values can appear here (like @me) instead; abandon parsing
    except ValueError:
        return

    try:
        # Ref: https://discordapp.com/developers/docs/reference#snowflakes
        timestamp = (snowflake >> 22) + 1420070400000
        worker_id = (snowflake & 0x3E0000) >> 17
        internal_process_id = (snowflake & 0x1F000) >> 17
        increment = snowflake & 0xFFF

    except Exception as e:
        log.exception(f'Exception parsing Discord snowflake: {e}')
        return

    node.hover = 'Discord Snowflakes are unique, time-based IDs. ' \
                 '<a href="https://discordapp.com/developers/docs/reference#snowflakes" target="_blank">[ref]</a>'

    unfurl.add_to_queue(
        data_type='epoch-milliseconds', key=None, value=timestamp, label=f'Timestamp:\n{timestamp}',
        hover='The first value in a Discord Snowflake is a timestamp associated with object creation',
        parent_id=node.node_id, incoming_edge_config=discord_edge)

    unfurl.add_to_queue(
        data_type='integer', key=None, value=worker_id, label=f'Worker ID: {worker_id}',
        hover='The second value in a Discord Snowflake is the internal worker ID',
        parent_id=node.node_id, incoming_edge_config=discord_edge)

    unfurl.add_to_queue(
        data_type='integer', key=None, value=internal_process_id, label=f'Process ID: {internal_process_id}',
        hover='The third value in a Discord Snowflake is the internal process ID',
        parent_id=node.node_id, incoming_edge_config=discord_edge)

    unfurl.add_to_queue(
        data_type='integer', key=None, value=increment, label=f'Increment: {increment}',
        hover='For every ID that is generated on that process, this number is incremented',
        parent_id=node.node_id, incoming_edge_config=discord_edge)


def run(unfurl, node):

    # Known patterns from main Discord site
    discord_domains = ['discordapp.com', 'discordapp.net']
    if any(discord_domain in unfurl.find_preceding_domain(node) for discord_domain in discord_domains):
        if node.data_type == 'url.path.segment':
            # Viewing a channel on a server
            # Ex: https://discordapp.com/channels/427876741990711298/551531058039095296
            if unfurl.check_sibling_nodes(node, data_type='url.path.segment', key=1, value='channels'):
                if node.key == 2:
                    unfurl.add_to_queue(
                        data_type='description', key=None, value=None, label='Server ID',
                        hover='The timestamp in the associated Discord Snowflake is the time of Server creation',
                        parent_id=node.node_id, incoming_edge_config=discord_edge)
                    parse_discord_snowflake(unfurl, node)

                elif node.key == 3:
                    unfurl.add_to_queue(
                        data_type='description', key=None, value=None, label='Channel ID',
                        hover='The timestamp in the associated Discord Snowflake is the time of Channel creation '
                              'on the given Server',
                        parent_id=node.node_id, incoming_edge_config=discord_edge)
                    parse_discord_snowflake(unfurl, node)

                # Linking to a specific message
                # Ex: https://discordapp.com/channels/427876741990711298/537760691302563843/643183730227281931
                elif node.key == 4:
                    unfurl.add_to_queue(
                        data_type='description', key=None, value=None, label='Message ID',
                        hover='The timestamp in the associated Discord Snowflake is the time the Message was sent',
                        parent_id=node.node_id, incoming_edge_config=discord_edge)
                    parse_discord_snowflake(unfurl, node)

            # File Attachment URLs
            # Ex: https://cdn.discordapp.com/attachments/622136585277931532/626893414490832918/asdf.png
            elif unfurl.check_sibling_nodes(node, data_type='url.path.segment', key=1, value='attachments'):
                if node.key == 2:
                    unfurl.add_to_queue(
                        data_type='description', key=None, value=None, label='Channel ID',
                        hover='The timestamp in the associated Discord Snowflake is the time of channel creation',
                        parent_id=node.node_id, incoming_edge_config=discord_edge)
                    parse_discord_snowflake(unfurl, node)

                elif node.key == 3:
                    unfurl.add_to_queue(
                        data_type='description', key=None, value=None, label='File ID',
                        hover='The timestamp in the associated Discord Snowflake is the time the attachment '
                              'was uploaded.',
                        parent_id=node.node_id, incoming_edge_config=discord_edge)
                    parse_discord_snowflake(unfurl, node)

                elif node.key == 4:
                    unfurl.add_to_queue(
                        data_type='description', key=None, value=None, label='Attachment File Name',
                        parent_id=node.node_id, incoming_edge_config=discord_edge)

            # Check if the node's value would correspond to a Snowflake with timestamp between 2015-02 and 2022-07
            elif unfurl.check_if_int_between(node.value, 15000000000000000, 1000000000000000001):
                parse_discord_snowflake(unfurl, node)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import requests
import json
import os


shortlink_edge = {
    'color': {
        'color': '#E7572C'
    },
    'title': 'URL Shortener Parser',
    'label': '🔗'
}
  

def expand_bitly_url(bitlink_id, api_key):
    # Ref: https://dev.bitly.com/v4/

    r = requests.post(
        'https://api-ssl.bitly.com/v4/expand',
        data=json.dumps({'bitlink_id': f'bit.ly/{bitlink_id.rstrip("/")}'}),
        headers={'Content-Type': 'application/json', 'Authorization': f'Bearer {api_key}'})

    if r.status_code == 200:
        return r.json()
    else:
        return {}


def expand_url_via_redirect_header(base_url, shortcode):
    r = requests.get(f'{base_url}{shortcode.rstrip("/")}', allow_redirects=False)

    if r.status_code in [301, 302]:
        return r.headers['Location']
    else:
        return {}


def run(unfurl, node):
    if not unfurl.remote_lookups:
        return

    # Linkedin has another method of URL shortening that's different from how most others do it; I can
    # refactor this in the future to be more flexible if I find more sites that operate this way, but for now
    # this works.
    if node.data_type == 'url.query.pair' and node.key == 'code':
        if 'linkedin.com' in unfurl.find_preceding_domain(node):
            expanded_url = expand_url_via_redirect_header('https://www.linkedin.com/slink?code=', node.value)
            if expanded_url:
                unfurl.add_to_queue(
                    data_type='url', key=None, value=expanded_url,
                    label=f'Expanded URL: {expanded_url}',
                    hover=f'Expanded URL, retrieved from linkedin.com via "Location" header',
                    parent_id=node.node_id, incoming_edge_config=shortlink_edge)
            return

    if not node.data_type == 'url.path':
        return

    bitly_domains = ['bit.ly', 'bitly.com', 'j.mp']
    if any(bitly_domain in unfurl.find_preceding_domain(node) for bitly_domain in bitly_domains):
        expanded_info = expand_bitly_url(node.value[1:], unfurl.api_keys.get('bitly', os.environ.get('bitly')))

        if not expanded_info:
            return

        node.hover = 'Bitly Short Links can be expanded via the Bitly API to show the ' \
                     '"long" URL and the creation time of the short-link.' \
                     '<a href="https://dev.bitly.com/v4/#operation/expandBitlink" ' \
                     'target="_blank">[ref]</a>'

        if expanded_info['created_at'].endswith('+0000'):
            expanded_info['created_at'] = expanded_info['created_at'][:-5]

        if expanded_info['created_at'][10] == 'T':
            expanded_info['created_at'] = f'{expanded_info["created_at"][:10]} {expanded_info["created_at"][11:]}'

        unfurl.add_to_queue(
            data_type='description', key=None, value=expanded_info['created_at'],
            label=f'Creation Time:\n{expanded_info["created_at"]}',
            hover='Short-link creation time, retrieved from Bitly API',
            parent_id=node.node_id, incoming_edge_config=shortlink_edge)

        unfurl.add_to_queue(
            data_type='url', key=None, value=expanded_info['long_url'],
            label=f'Expanded URL: {expanded_info["long_url"]}', hover='Expanded URL, retrieved from Bitly API',
            parent_id=node.node_id, incoming_edge_config=shortlink_edge)

        return

    redirect_expands = [
        {'domain': 'bit.do', 'base_url': 'https://bit.do/'},
        {'domain': 'buff.ly', 'base_url': 'https://buff.ly/'},
        {'domain': 'cutt.ly', 'base_url': 'https://cutt.ly/'},
        {'domain': 'db.tt', 'base_url': 'https://db.tt/'},
        {'domain': 'dlvr.it', 'base_url': 'https://dlvr.it/'},
        {'domain': 'fb.me', 'base_url': 'https://fb.me/'},
        {'domain': 'goo.gl', 'base_url': 'https://goo.gl/'},
        {'domain': 'ift.tt', 'base_url': 'https://ift.tt/'},
        {'domain': 'is.gd', 'base_url': 'https://is.gd/'},
        {'domain': 'lc.chat', 'base_url': 'https://lc.chat/'},
        {'domain': 'lnkd.in', 'base_url': 'https://www.linkedin.com/slink?code='},
        {'domain': 'nyti.ms', 'base_url': 'https://nyti.ms/'},
        {'domain': 'ow.ly', 'base_url': 'http://ow.ly/'},
        {'domain': 'reut.rs', 'base_url': 'https://reut.rs/'},
        {'domain': 'sansurl.com', 'base_url': 'https://sansurl.com/'},
        {'domain': 'snip.ly', 'base_url': 'https://snip.ly/'},
        {'domain': 't.co', 'base_url': 'https://t.co/'},
        {'domain': 't.ly', 'base_url': 'https://t.ly/'},
        {'domain': 'tr.im', 'base_url': 'https://tr.im/'},
        {'domain': 'trib.al', 'base_url': 'https://trib.al/'},
        {'domain': 'tinyurl.com', 'base_url': 'https://tinyurl.com/'},
        {'domain': 'urlwee.com', 'base_url': 'https://urlwee.com/'},
        {'domain': 'urlzs.com', 'base_url': 'https://urlzs.com/'},
        {'domain': 'x.co', 'base_url': 'https://x.co/'},
    ]

    preceding_domain = unfurl.find_preceding_domain(node)
    for redirect_expand in redirect_expands:
        if redirect_expand['domain'] == preceding_domain:
            expanded_url = expand_url_via_redirect_header(redirect_expand['base_url'], node.value[1:])
            if expanded_url:
                unfurl.add_to_queue(
                    data_type='url', key=None, value=expanded_url,
                    label=f'Expanded URL: {expanded_url}',
                    hover=f'Expanded URL, retrieved from {redirect_expand["domain"]} via "Location" header',
                    parent_id=node.node_id, incoming_edge_config=shortlink_edge)
            return

    # Get the list of "known" URL shortener domains from MISP; many of these seem to be deprecated.
    # Try to expand the shortlink via a 301/302 Location header; if the site uses something like a meta refresh,
    # this won't work.
    misp_shortener_domains = unfurl.known_domain_lists['List of known URL Shorteners domains'].list
    if preceding_domain in misp_shortener_domains:
        expanded_url = expand_url_via_redirect_header(f'https://{preceding_domain}/', node.value[1:])
        if expanded_url:
            unfurl.add_to_queue(
                data_type='url', key=None, value=expanded_url,
                label=f'Expanded URL: {expanded_url}',
                hover=f'Expanded URL, retrieved from {preceding_domain} via "Location" header',
                parent_id=node.node_id, incoming_edge_config=shortlink_edge)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import datetime
import struct
import pycountry
from unfurl.parsers.proto.google_search_pb2 import Ved
from google.protobuf import json_format

import logging
log = logging.getLogger(__name__)

google_edge = {
    'color': {
        'color': 'green'
    },
    'title': 'Google-related Parsing Functions',
    'label': 'G'
}

known_sources = {
    'hp': 'Home Page',
    'serp': 'Search Engine Results Page',
    'img': 'Image Search',
    'psy-ab': 'Address Bar',
    'hangouts': 'Hangouts Chat'
}


def decode_varint(source):
    result = 0
    number_of_bytes = 0
    for read in source:
        result |= ((read & 0x7F) << (number_of_bytes * 7))
        number_of_bytes += 1
        if (read & 0x80) != 0x80:
            return result, number_of_bytes


def split_exactly(to_split, delimiter, times):
    x = to_split.split(delimiter, times)
    while len(x) < times+1:
        x.append(None)
    return x


def parse_aqs(aqs: str) -> dict:
    parsed = {}
    parsed['Device Type'], parsed['clicked_suggestion'], remainder = aqs.split('.', 2)
    autocompletion_str, timing_str = split_exactly(remainder, '.', 1)
    parsed['autocompletion_entries'] = autocompletion_str.split('j')

    if timing_str:
        parsed['query_formulation_time'], parsed['field_trial_triggered'], parsed['page_classification'] = \
            timing_str.split('j', 2)

    return parsed


def parse_ei(ei):
    """Parse ei parameters from Google searches.

    Based on work by:
      Adrian Leong (cheeky4n6monkey@gmail.com) -
        https://github.com/cheeky4n6monkey/4n6-scripts/blob/master/google-ei-time.py
      Kevin Jones - https://deedpolloffice.com/blog/articles/decoding-ei-parameter
    """

    decoded = base64.urlsafe_b64decode(ei)

    # grab 1st 4 bytes and treat as LE unsigned int
    timestamp = struct.unpack('<i', decoded[0:4])[0]
    parsed = [timestamp]

    varint_offset = 4  # First 4 (0-3) bytes are the timestamp
    for _ in [1, 2, 3]:
        try:
            value, bytes_used = decode_varint(decoded[varint_offset:])
            parsed.append(value)
            varint_offset += bytes_used
        except TypeError as e:
            log.warning(f'Unable to decode varint from {decoded}: {e}')
            return

    return parsed


def parse_rlz(rlz_string):
    # Reference: https://github.com/rogerta/rlz/blob/wiki/HowToReadAnRlzString.md

    rlz = {}
    assert rlz_string[0] == '1', 'Only v1 of RLZ is known'

    rlz['version'] = {
            'data_type': 'google.rlz.version',
            'key': 'RLZ version',
            'value': rlz_string[0],
            'hover': 'Only v1 of RLZ is known'}

    rlz['ap'] = {
        'data_type': 'google.rlz.ap',
        'key': 'Application',
        'value': rlz_string[1:3],
        'hover': 'Application (or "access point") used to do the Google Search'}

    rlz['brand_code'] = {
        'data_type': 'google.rlz.brand_code',
        'key': 'Brand Code',
        'value': rlz_string[3:7],
        'hover': 'The brand code identifies the distribution channel (it may be a partner or internal marketing). <br>'
                 'This correlates to how the user got the software (ie. they downloaded it by itself vs. it came <br>'
                 'pre-installed on their new computer vs. it came with a partner\'s software.'}

    cannibal_string = rlz_string[7]
    if cannibal_string == 'c':
        cannibal_value = 'Yes'
    elif cannibal_string == '_':
        cannibal_value = 'No'
    else:
        raise ValueError(f'Invalid "cannibal" value in RLZ string: {cannibal_string}')

    rlz['cannibal'] = {
        'data_type': 'google.rlz.cannibal',
        'key': 'Cannibalized',
        'value': cannibal_value,
        'hover': '"Cannibal" tells if the library has evidence that the user was a user prior <br>'
                 'to installing the software, and thus "cannibalized" a previous installation.'}

    rlz_pointer = 8

    # Examples of longer RLZ values (2- and 5-char lang codes):
    #   1C1CHBF_en-GBGB901GB901
    #   1C1GCEU_enUS820US820
    if len(rlz_string) >= 11:
        # If a dash is in this position, it means it is a 5 char lang code, not the 2 char code
        if rlz_string[10] == '-':
            language_code = rlz_string[rlz_pointer:rlz_pointer+5]
            rlz_pointer += 5
            # langcodes was having install issues on macOS; not using it for now in
            # order to not complicate Unfurl's install. Pycountry's languages isn't
            # as good (only alpha_2 and alpha_3) but better than nothing for now.
            # Old implementation:
            # language_name = langcodes.Language.get(language_code).language_name()
            language_name = pycountry.languages.get(alpha_2=language_code[:2]).name

        else:
            language_code = rlz_string[rlz_pointer:rlz_pointer+2]
            # language_name = langcodes.Language.get(language_code).language_name()
            language_name = pycountry.languages.get(alpha_2=language_code).name
            rlz_pointer += 2

    # Example of RLZ value without cohorts
    #   1C1GCEV_en
    elif len(rlz_string) == 10:
        language_code = rlz_string[rlz_pointer:rlz_pointer + 2]
        # language_name = langcodes.Language.get(language_code).language_name()
        language_name = pycountry.languages.get(alpha_2=language_code).name
        rlz_pointer += 2

    else:
        raise ValueError('Couldn\'t parse RLZ language code')

    rlz['language'] = {
        'data_type': 'google.rlz.language',
        'key': 'Language',
        'value': f'{language_name} ({language_code})',
        'hover': 'The two- (en) or five-character (zh-CN) language code of the application. <br>'
                 'Valid values depend on the specific app.'}

    # The cohorts are optional, so if this is the end of the RLZ string, return what we've parsed
    if not len(rlz_string) >= rlz_pointer+5:
        return rlz

    install_cohort = rlz_string[rlz_pointer:rlz_pointer+5]
    rlz_pointer += 5

    install_country_code = install_cohort[:2]
    install_country = pycountry.countries.get(alpha_2=install_country_code).name
    install_date = datetime.timedelta(weeks=int(install_cohort[2:])) + datetime.date(year=2003, month=2, day=3)

    rlz['install_cohort'] = {
        'data_type': 'google.rlz.install_cohort',
        'key': 'Install Cohort',
        'value': f'Installed in {install_country} the week of {install_date}',
        'hover': 'The country and week of the user\'s installation event. <br>'
                 'Country is determined by the server, using IP address. <br>'
                 'Week is measured as number of weeks since Feb 3, 2003.'}

    # Both cohorts do not need to be present, so if the 2nd one (search) is missing, return what we have
    # Example: 1CAGZLV_enGB898
    if not len(rlz_string) >= rlz_pointer + 5:
        return rlz

    search_cohort = rlz_string[rlz_pointer:rlz_pointer + 5]

    search_country_code = search_cohort[:2]
    search_country = pycountry.countries.get(alpha_2=search_country_code).name
    search_date = datetime.timedelta(weeks=int(search_cohort[2:])) + datetime.date(year=2003, month=2, day=3)

    rlz['search_cohort'] = {
        'data_type': 'google.rlz.search_cohort',
        'key': 'Search Cohort',
        'value': f'First search was in {search_country} the week of {search_date}',
        'hover': 'The country and week of the user\'s first search. <br>'
                 'Country is determined by the server, using IP address. <br>'
                 'Week is measured as number of weeks since Feb 3, 2003.'}

    return rlz


def run(unfurl, node):
    if node.data_type == 'url.query.pair':
        if 'google' in unfurl.find_preceding_domain(node):
            if node.key == 'biw':
                unfurl.add_to_queue(
                    data_type='google.biw', key='Browser Width', value=node.value,
                    label=f'Browser width: {node.value}px',
                    hover='Inner width of the browser window',
                    parent_id=node.node_id, incoming_edge_config=google_edge)

            elif node.key == 'bih':
                biw_node = unfurl.check_sibling_nodes(node, data_type='url.query.pair', key='biw', return_node=True)
                if biw_node:
                    unfurl.add_to_queue(
                        data_type='descriptor', key=None,
                        value=f'Size of browser window: {biw_node.value}x{node.value} pixels',
                        hover='The size of the "content area" in browser window',
                        parent_id=[node.node_id, biw_node.node_id], incoming_edge_config=google_edge)

                unfurl.add_to_queue(
                    data_type='google.bih', key='Browser Height', value=node.value,
                    label=f'Browser height: {node.value}px',
                    hover='Inner height of the browser window',
                    parent_id=node.node_id, incoming_edge_config=google_edge)

            elif node.key == 'aqs':
                parsed_aqs = parse_aqs(node.value)
                for key, value in parsed_aqs.items():
                    if key == 'autocompletion_entries':
                        repeated_count_offset = 0
                        for index, match in enumerate(value):
                            adjusted_start = index + repeated_count_offset
                            if 'l' in match:
                                _, repeated_count = split_exactly(match, 'l', 1)
                                repeated_count = int(repeated_count) - 1
                                repeated_count_offset += repeated_count
                                unfurl.add_to_queue(
                                    data_type='google.aqs.ac_match',
                                    key=f'Autocomplete Matches ({adjusted_start}-{adjusted_start + repeated_count})',
                                    value=match, parent_id=node.node_id, incoming_edge_config=google_edge)
                            else:
                                unfurl.add_to_queue(
                                    data_type='google.aqs.ac_match',
                                    key=f'Autocomplete Match ({adjusted_start})',
                                    value=match, parent_id=node.node_id, incoming_edge_config=google_edge)

                    elif key == 'page_classification':
                        unfurl.add_to_queue(
                            data_type='google.aqs.page_classification', key='Page Classification', value=value,
                            hover='The type of page currently displayed when the user used the omnibox',
                            parent_id=node.node_id, incoming_edge_config=google_edge)

                    elif key == 'query_formulation_time':
                        unfurl.add_to_queue(
                            data_type='google.aqs.query_formulation_time', key='Query Formulation Time',
                            value=f'{int(value)/1000} seconds',
                            hover='Query formulation time (time from when the user first typed a character into the '
                                  'omnibox to when the user selected a query)',
                            parent_id=node.node_id, incoming_edge_config=google_edge)

                    elif key == 'field_trial_triggered':
                        unfurl.add_to_queue(
                            data_type='google.aqs.field_trial_triggered', key='Field Trial Triggered',
                            value=str(bool(int(value))),  # converts the '0' or '1' (strings) into True or False strings
                            hover='Whether a field trial (of zero suggest or search provider) was triggered in the '
                                  'session',
                            parent_id=node.node_id, incoming_edge_config=google_edge)

                    elif key == 'clicked_suggestion':
                        if value:
                            unfurl.add_to_queue(
                                data_type='google.aqs.clicked_suggestion', key='Clicked Suggestion',
                                value=value, hover='The index (starting at 0) of the accepted suggestion',
                                parent_id=node.node_id, incoming_edge_config=google_edge)
                        else:
                            unfurl.add_to_queue(
                                data_type='google.aqs.clicked_suggestion', key='Clicked Suggestion',
                                value='None', hover='The user did not click one of the offered suggestions',
                                parent_id=node.node_id, incoming_edge_config=google_edge)
                    else:
                        unfurl.add_to_queue(
                            data_type='google.aqs', key=key, value=value,
                            parent_id=node.node_id, incoming_edge_config=google_edge)

            elif node.key == 'ei':
                padded_value = unfurl.add_b64_padding(node.value)
                if not padded_value:
                    return

                try:
                    parsed_ei = parse_ei(padded_value)
                except Exception as e:
                    log.exception(f'Exception running parse_ei() on {padded_value}: {e}')
                    return

                if not parsed_ei:
                    return

                node.hover = 'The \'<b>ei</b>\' parameter is base64-encoded and contains four values. ' \
                             '<br>The first two are thought to be the timestamp of when the session started ' \
                             '<br>(first value is full seconds, second is microsecond component)' \
                             '<br>This may be seconds (but could be days!) before the URL was generated.' \
                             '<br><br>References:<ul>' \
                             '<li><a href="https://deedpolloffice.com/blog/articles/decoding-ei-parameter" ' \
                             'target="_blank">Kevin Jones: How to decode the ei parameter in Google search</a></li>' \
                             '<li><a href="http://cheeky4n6monkey.blogspot.com/2014/10/google-eid.html" ' \
                             'target="_blank">Cheeky4n6Monkey: Google-ei\'d ?!</a></li>' \
                             '<li><a href="https://github.com/obsidianforensics/unfurl/issues/56" ' \
                             'target="_blank">Rasmus-Riis: Search Experiment with ei parameter</a></li>'\
                             '<li>Adam Mazack: noticed the 2nd ei value contained fractional seconds ' \
                             '<br> that match the ved</li></ul>'
                node.extra_options = {'widthConstraint': {'maximum': 300}}

                assert len(parsed_ei) == 4, \
                    f'There should be 4 decoded ei values, but we have {len(parsed_ei)}!'

                unfurl.add_to_queue(
                    data_type='google.ei', key=0, value=parsed_ei[0],
                    label=f'ei-0: {parsed_ei[0]}',
                    hover='The first <b>ei</b> value is thought to be part of the timestamp'
                          ' (seconds) of when the session began.',
                    parent_id=node.node_id, incoming_edge_config=google_edge)

                unfurl.add_to_queue(
                    data_type='google.ei', key=1, value=parsed_ei[1],
                    label=f'ei-1: {parsed_ei[1]}',
                    hover='The second <b>ei</b> value is thought to be part of the timestamp'
                          ' (fractional seconds) of when the session began.',
                    parent_id=node.node_id, incoming_edge_config=google_edge)

                unfurl.add_to_queue(
                    data_type='google.ei', key=2, value=parsed_ei[2],
                    label=f'ei-2: {parsed_ei[2]}',
                    hover='The meaning of the third <b>ei</b> value is not known.',
                    parent_id=node.node_id, incoming_edge_config=google_edge)

                unfurl.add_to_queue(
                    data_type='google.ei', key=3, value=parsed_ei[3],
                    label=f'ei-3: {parsed_ei[3]}',
                    hover='The meaning of the fourth <b>ei</b> value is not known, '
                          'but it matches the <b>ved</b> "13-3" value.',
                    parent_id=node.node_id, incoming_edge_config=google_edge)

            elif node.key == 'gs_l':
                node.hover = 'The <b>gs_l</b> parameter contains multiple values, delimited by <b>.</b>. <br>' \
                             'We only know the meaning of 7 (of the 26+) parameters.'
                node.extra_options = {'widthConstraint': {'maximum': 1400}}

                params = node.value.split('.')

                known_params = [0, 1, 2, 4, 5, 7, 8, 26]
                for known_param in known_params:
                    if len(params) > known_param and len(params[known_param]) > 0:
                        unfurl.add_to_queue(
                            data_type='google.gs_l', key=str(known_param), value=params[known_param],
                            label=f'Parameter {str(known_param)}: {params[known_param]}',
                            parent_id=node.node_id, incoming_edge_config=google_edge)

            elif node.key == 'oq':
                unfurl.add_to_queue(
                    data_type='google.oq', key=None, value=f'"Original" Search Query: {node.value}',
                    hover='Original terms entered by the user; auto-complete or suggestions <br>'
                          'may have been used to reach the actual search terms (in <b>q</b>)',
                    parent_id=node.node_id, incoming_edge_config=google_edge)

            elif node.key == 'q':
                unfurl.add_to_queue(
                    data_type='google.q', key=None, value=f'Search Query: {node.value}',
                    hover='Terms used in the Google search', parent_id=node.node_id, incoming_edge_config=google_edge)

            elif node.key == 'source':
                if node.value in known_sources.keys():
                    unfurl.add_to_queue(
                        data_type='google.source', key=None, value=f'Source: {known_sources[node.value]}',
                        hover='Source of the Google search', parent_id=node.node_id, incoming_edge_config=google_edge)

            elif node.key == 'start':
                unfurl.add_to_queue(
                    data_type='google.start', key=None, value=f'Starting Result: {node.value}',
                    hover='Google search by default shows 10 results per page; higher <br>'
                          '"start" values may indicate browsing more subsequent results pages.',
                    parent_id=node.node_id, incoming_edge_config=google_edge)

            elif node.key == 'sxsrf':
                node.extra_options = {'widthConstraint': {'maximum': 400}}
                sxs_0, sxs_1 = node.value.split(':', 1)
                unfurl.add_to_queue(data_type='google.sxsrf', key=1, value=sxs_0,
                                    parent_id=node.node_id, incoming_edge_config=google_edge)
                unfurl.add_to_queue(
                    data_type='epoch-milliseconds', key=2, value=sxs_1,
                    hover='The <b>sxsrf</b> parameter contains a timestamp, believed<br>'
                          ' to correspond to the previous page load. <br><br>Refernces:<ul><li>'
                          '<a href="https://twitter.com/phillmoore/status/1169846359509233664" target="_blank">'
                          'Phill Moore on Twitter</a></li></ul>', parent_id=node.node_id,
                    incoming_edge_config=google_edge)

            elif node.key == 'tbm':
                tbm_mappings = {
                    'bks': 'Google Books',
                    'fin': 'Google Finance',
                    'flm': 'Google Flights',
                    'isch': 'Google Images',
                    'nws': 'Google News',
                    'shop': 'Google Shopping',
                    'vid': 'Google Videos',
                }

                value = tbm_mappings.get(node.value, 'Unknown')
                unfurl.add_to_queue(
                    data_type='google.tbm', key=None, value=f'Search Type: {value}',
                    hover='Google Search Type', parent_id=node.node_id, incoming_edge_config=google_edge)

            elif node.key == 'uule':
                # https://moz.com/ugc/geolocation-the-ultimate-tip-to-emulate-local-search
                location_string = base64.b64decode(unfurl.add_b64_padding(node.value[10:]))
                unfurl.add_to_queue(
                    data_type='google.uule', key=None, value=location_string, label=location_string,
                    parent_id=node.node_id, incoming_edge_config=google_edge)

            elif node.key == 'rlz':
                unfurl.add_to_queue(
                    data_type='descriptor', key=None,
                    value='RLZ used for grouping promotion event signals and anonymous user cohorts',
                    parent_id=node.node_id, incoming_edge_config=google_edge)

                try:
                    rlz = parse_rlz(node.value)

                    for component in rlz.values():
                        unfurl.add_to_queue(
                            data_type=component['data_type'], key=component['key'],
                            value=component['value'], hover=component['hover'],
                            parent_id=node.node_id, incoming_edge_config=google_edge)

                except ValueError as e:
                    print(f'Exception parsing RLZ: {e}')

            elif node.key == 'ved':
                node.extra_options = {'widthConstraint': {'maximum': 400}}
                known_ved_descriptions = {
                    'linkIndex':
                        'Unique index for each link on the search page. <br>The higher the number, the farther'
                        ' down the page the link is. <br>This should always be present. '
                        '<a href="https://deedpolloffice.com/blog/articles/decoding-ved-parameter" '
                        'target="_blank">[ref]</a>',
                    'linkType':
                        'The type of link that was clicked on; there are thousands. <br>This should always be '
                        'present. <a href="https://deedpolloffice.com/blog/articles/decoding-ved-parameter" '
                        'target="_blank">[ref]</a>',
                    'subResultPosition':
                        'The position of the link, if it was inside a "group" (like an adword or<br> '
                        'knowledge graph). This starts at 0 and counts up.'
                        '<a href="https://deedpolloffice.com/blog/articles/decoding-ved-parameter" '
                        'target="_blank">[ref]</a>',
                    'resultPosition':
                        'The position of the result on the page. The higher the number, <br>the farther '
                        'down the page the result is. '
                        '<a href="https://deedpolloffice.com/blog/articles/decoding-ved-parameter" '
                        'target="_blank">[ref]</a>',
                    'resultsStart':
                        'The starting position of the first result on the page. <br>On page 2, it will be '
                        '10; and on page 3 it will be 20, <br>and so on. On page 1, the value isn’t '
                        'present (but implicitly, this means a value of 0).'
                        '<a href="https://deedpolloffice.com/blog/articles/decoding-ved-parameter" '
                        'target="_blank">[ref]</a>',
                }
                assert node.value[0] in ['0', '2'], 'The ved parameter should start with 0 or 2'
                encoded_ved = node.value[1:]
                padded_ved = unfurl.add_b64_padding(encoded_ved)
                if not padded_ved:
                    return
                try:
                    encoded_ved = base64.urlsafe_b64decode(padded_ved)
                    ved = Ved().FromString(encoded_ved)
                except Exception as e:
                    log.warning(f'Unable to parse ved from {padded_ved}: {e}')
                    return

                ved_dict = json_format.MessageToDict(ved)
                for key, value in ved_dict.items():
                    if key == 'v13Outer':
                        assert isinstance(value, dict), 'ved-13-Outer should be a dict'
                        assert isinstance(value['v13Inner'], dict), 'ved-13-Inner should also be a dict'
                        v13_timestamp = value['v13Inner'].get('timestamp')
                        if v13_timestamp:
                            unfurl.add_to_queue(
                                data_type='epoch-microseconds', key='Timestamp', value=v13_timestamp,
                                parent_id=node.node_id, incoming_edge_config=google_edge,
                                hover='The ved parameter contains a timestamp, believed to <br>correspond to around '
                                      'when the page loaded.')

                        v13_unknown_2 = value['v13Inner'].get('v132')
                        if v13_unknown_2:
                            unfurl.add_to_queue(
                                data_type='google.ved', key='13-2', value=v13_unknown_2,
                                parent_id=node.node_id, incoming_edge_config=google_edge,
                                hover='Inside the ved parameter, the meanings of<br> 13-2 and 13-3 are not known.')

                        v13_unknown_3 = value['v13Inner'].get('v133')
                        if v13_unknown_3:
                            unfurl.add_to_queue(
                                data_type='google.ved', key='13-3', value=v13_unknown_3,
                                parent_id=node.node_id, incoming_edge_config=google_edge,
                                hover='Inside the ved parameter, the meanings of 13-2 and 13-3 are not known, <br>'
                                      'but the values in <b>ved 13-3</b> and <b>ei-3</b> match.')

                    elif key == 'v15':
                        assert isinstance(value, dict), 'ved-15 should be a dict'
                        v15_1 = value.get('v151')
                        v15_2 = value.get('v152')

                        if v15_1:
                            unfurl.add_to_queue(
                                data_type='google.ved', key='15-1', value=v15_1,
                                parent_id=node.node_id, incoming_edge_config=google_edge,
                                hover='Inside the ved parameter, the meanings of<br> 15-1 and 15-2 are not known.')
                        if v15_2:
                            unfurl.add_to_queue(
                                data_type='google.ved', key='15-2', value=v15_2,
                                parent_id=node.node_id, incoming_edge_config=google_edge,
                                hover='Inside the ved parameter, the meanings of<br> 15-1 and 15-2 are not known.')

                    else:
                        unfurl.add_to_queue(
                            data_type='google.ved', key=key, value=value, label=f'{key}: {value}',
                            hover=known_ved_descriptions.get(key, ''),
                            parent_id=node.node_id, incoming_edge_config=google_edge)

    if node.data_type == 'google.aqs.ac_match':
        ac_types = {
            0: 'Suggested Search',
            # A suggested search (with the default engine) query that doesn't fall into one of the
            # more specific suggestion categories below.
            5: 'Suggested URL',
            6: 'Calculator',
            33: 'Suggested Search - Tail',  # A suggested search to complete the tail of the query.
            35: 'Personalized Suggested Search',  # A personalized suggested search.
            44: 'Personalized Suggested Search for a Google+ profile',
            46: 'Suggested Entity Search',  # A suggested search for an entity.
            69: 'Native Chrome Suggestion',
            171: 'Tile Suggestion'
        }

        ac_subtypes = {
            39: 'Personal',
            57: 'Omnibox Echo Search',
            58: 'Omnibox Echo URL',
            59: 'Omnibox History Search',
            60: 'Omnibox History URL',
            61: 'Omnibox History Title',
            62: 'Omnibox History Body',
            63: 'Omnibox History Keyword',
            64: 'Omnibox Other',  # This value indicates a native chrome suggestion with no named subtype
            65: 'Omnibox Bookmark Title',
            66: 'URL-based',
            176: 'Clipboard Text',
            177: 'Clipboard URL',
            271: 'Suggest 2G Lite',
            327: 'Clipboard Image',
            362: 'Zero Prefix',
            450: 'Zero Prefix Local History',
            451: 'Zero Prefix Local Frequent URLs'
        }

        entry, count = split_exactly(node.value, 'l', 1)
        ac_type, subtype_str = split_exactly(entry, 'i', 1)

        unfurl.add_to_queue(
            data_type='descriptor', key='Type', value=ac_types.get(int(ac_type), f'Unknown ({ac_type})'),
            parent_id=node.node_id, incoming_edge_config=google_edge)

        if subtype_str:
            for subtype in subtype_str.split('i'):
                unfurl.add_to_queue(
                    data_type='descriptor', key='Subtype', value=ac_subtypes.get(int(subtype), f'Unknown ({subtype})'),
                    parent_id=node.node_id, incoming_edge_config=google_edge)

        if count:
            unfurl.add_to_queue(
                data_type='descriptor', key='Count', value=count,
                parent_id=node.node_id, incoming_edge_config=google_edge)

    elif node.data_type == 'google.aqs.page_classification':
        # Source: https://source.chromium.org/chromium/chromium/src/+/main:third_party/
        #         metrics_proto/omnibox_event.proto;l=97
        classifications = {
            0: {
                'name': 'Invalid Spec',
                'description': 'An invalid URL; shouldn\'t happen'
            },

            1: {
                'name': 'New Tab Page',
                'description': 'The New Tab Page (chrome://newtab/). For modern versions of Chrome, this is only '
                               'reported when an extension is replacing the new tab page. Otherwise, new tab page '
                               'interactions will be as one of the other "New Tab Page" types. For old versions of '
                               'Chrome, this was reported for the default New Tab Page.',
            },

            2: {
                'name': 'Blank Page',
                'description': 'about:blank'
            },

            3: {
                'name': 'Home Page',
                'description': 'The user\'s home page. Note that if the home page is set to any of the new tab page '
                               'versions or to about:blank, then we\'ll classify the page into those categories, '
                               'not "Home Page".'
            },

            4: {
                'name': 'Other',
                'description': 'The catch-all entry of everything not included somewhere else on this list.'
            },

            5: {
                'name': '(Obsolete) Instant New Tab Page',
                'description': 'The instant new tab page enum value was deprecated on August 2, 2013.'
            },

            6: {
                'name': 'Search Result Page (with search term replacement)',
                'description': 'The user is on a search result page that does search term replacement. This means the '
                               'search terms are shown in the omnibox instead of the URL. In other words: Query in '
                               'Omnibox is Active for this SRP.'
            },

            7: {
                'name': 'Instant New Tab Page (with omnibox as starting focus)',
                'description': 'The new tab page in which this omnibox interaction first started with the user having '
                               'focus in the omnibox.'
            },

            8: {
                'name': 'Instant New Tab Page (with fakebox as starting focus)',
                'description': 'The new tab page in which this omnibox interaction first started with the user having '
                               'focus in the fakebox. Note that this started being replaced by NTP_REALBOX in Aug 2020 '
                               'and will eventually be obsolete.'
            },

            9: {
                'name': 'Search Results Page (no search term replacement)',
                'description': 'The user is on a search result page that does not do search term replacement. This '
                               'means the URL of the SRP is shown in the omnibox. In other words: Query in Omnibox is '
                               'Inactive for this SRP.'
            },

            10: {
                'name': 'Home Screen',
                'description': 'The user is on the home screen.'
            },

            11: {
                'name': 'Search App',
                'description': 'The user is in the search app.'
            },

            12: {
                'name': 'Maps App',
                'description': 'The user is in the maps app.'
            },

            13: {
                'name': 'Search Button',
                'description': 'This omnibox interaction started with the user tapping the search button.'
            },

            14: {
                'name': 'ChromeOS App List',
                'description': 'This interaction started with the user focusing or typing in the search box of the '
                               'ChromeOS app list (a.k.a., launcher).'
            },

            15: {
                'name': 'New Tab Page (realbox)',
                'description': 'The new tab page in which this omnibox interaction started with the user having focus '
                               'in the realbox.'
            },

            16: {
                'name': 'Android Search Widget',
                'description': 'Android\'s Search Widget started directly from Launcher.'
            },

            17: {
                'name': 'Android Start surface homepage',
                'description': 'Android\'s Start surface homepage.'
            },

            18: {
                'name': 'Android Start surface New Tab',
                'description': 'New Tab with Omnibox focused when Android\'s start surface finale is enabled.'
            },

            19: {
                'name': 'Android Shortcuts Widget',
                'description': 'Android\'s improved Search Widget with new suggestions'
            }
        }

        known = classifications.get(int(node.value), None)
        if not known:
            return

        unfurl.add_to_queue(
            data_type='descriptor', key=None, value=f'{node.value}: {known["name"]}', hover=known['description'],
            parent_id=node.node_id, incoming_edge_config=google_edge)

    elif node.data_type == 'google.gs_l':
        known_values = {
            '0': known_sources,
            '1': {
                '1': 'mouse-click on suggestion',
                '3': 'keyboard [enter] key',
                '10': 'Google Instant Search (untested)'
            }
        }
        if node.key == '0':
            if node.value in known_values[node.key].keys():
                unfurl.add_to_queue(
                    data_type='descriptor', key=None, value=known_values[node.key][node.value],
                    label=f'Searcher came from {known_values[node.key][node.value]}',
                    parent_id=node.node_id, incoming_edge_config=google_edge)

        if node.key == '1':
            if node.value in known_values[node.key].keys():
                unfurl.add_to_queue(
                    data_type='descriptor', key=None, value=known_values[node.key][node.value],
                    label=f'Searcher selected search terms by {known_values[node.key][node.value]}',
                    parent_id=node.node_id, incoming_edge_config=google_edge)

        if node.key == '2':
            unfurl.add_to_queue(
                data_type='descriptor', key=None, value=node.value,
                label=f'Selected option {node.value} from suggestion list',
                parent_id=node.node_id, incoming_edge_config=google_edge)

        if node.key == '4':
            if node.value == '0':
                param_text = 'User clicked on a suggestion'
            else:
                param_text = f'{str(float(node.value) / 1000)} seconds before the user selected the search box'

            unfurl.add_to_queue(
                data_type='descriptor', key=None, value=node.value, label=param_text,
                parent_id=node.node_id, incoming_edge_config=google_edge)

        if node.key == '5':
            if node.value == '0':
                param_text = 'User clicked on a suggestion'
            else:
                param_text = f'{str(float(node.value) / 1000)} seconds on page before the user completed typing.'

            unfurl.add_to_queue(
                data_type='descriptor', key=None, value=node.value, label=param_text,
                parent_id=node.node_id, incoming_edge_config=google_edge)

        if node.key == '7':
            if node.value != '0':
                param_text = f'User spent {str(float(node.value) / 1000)} seconds on page in total.'
                unfurl.add_to_queue(
                    data_type='descriptor', key=None, value=node.value, label=param_text,
                    parent_id=node.node_id, incoming_edge_config=google_edge)

        if node.key == '8':
            if node.value == '0':
                param_text = 'No characters typed'
            else:
                param_text = f'{node.value} keys pressed while user was typing.'

            unfurl.add_to_queue(
                data_type='descriptor', key=None, value=node.value, label=param_text,
                hover='If the search was done in Chrome, this value may be +1.',
                parent_id=node.node_id, incoming_edge_config=google_edge)

        if node.key == '26':
            if node.value == '1':
                unfurl.add_to_queue(
                    data_type='descriptor', key=None, value=node.value, label='User typed in query',
                    parent_id=node.node_id, incoming_edge_config=google_edge)

    elif node.data_type == 'google.ved':
        known_link_types = {
            22: 'web',
            152: 'blog search result',
            232: 'book or patent search result',
            235: 'book or patent search result thumbnail',
            244: 'image result in standard SERPs',
            245: 'image search result in basic (non-javascript) image search, or image result in universal search',
            288: 'local search result',
            295: 'news result thumbnail',
            297: 'news result',
            300: 'more results link (listed mainly for Q&A websites)',
            311: 'video result',
            312: 'video result thumbnail',
            338: 'one-line sitelink',
            371: 'shopping search result',
            429: 'image search result [probably not in use any more]',
            586: '"Jump to" link',
            612: 'map search result website link',
            646: 'map search result thumbnail',
            706: 'adword sitelink',
            745: 'breadcrumb',
            1107: 'patent result "Overview" / "Related" / "Discuss" link',
            1140: 'book search result author link',
            1146: 'normal result thumbnail (e.g. for an application, recipe, etc.)',
            1150: 'normal result thumbnail (e.g. for an application, recipe, etc.)',
            1455: 'local search result marker pin icon',
            1532: 'news sub-result (i.e. the same story from a different site)',
            1617: 'adword (i.e. sponsored search result)',
            1701: 'map search result',
            1732: 'knowledge graph repeated sub-link (e.g. football team squad players, album track listings)',
            1907: 'sponsored shopping result thumbnail (in right-hand column of universal search results)',
            1908: 'sponsored shopping result (in right-hand column of universal search results)',
            1986: 'sponsored shopping result thumbnail (in main column of universal search results)',
            1987: 'sponsored shopping result (in main column of universal search results)',
            2060: 'sitelink',
            2237: 'news result video thumbnail',
            2459: 'knowledge graph link',
            2847: 'authorship "by [author]" link',
            2937: 'authorship thumbnail link',
            3588: 'image search result (thumbnail)',
            3596: 'image search result preview "View image" link',
            3597: 'image search result preview thumbnail',
            3598: 'image search result preview title link',
            3599: 'image search result preview "Visit page" link',
            3724: 'image search result preview grey website link underneath title',
            3836: 'knowledge graph main image',
            5077: 'in-depth article result',
            5078: 'in-depth article result thumbnail',
            5158: 'adword one-line sitelink',
            5497: 'dictionary definition link'
        }

        if node.key == 'linkType':
            if known_link_types.get(node.value):
                unfurl.add_to_queue(
                    data_type='descriptor', key=None, value=node.value, label=known_link_types.get(node.value),
                    hover='There are tens of thousands of these values; the \'known\'<br> ones in Unfurl are based on '
                          '<a href="https://github.com/beschulz/ved-decoder" target="_blank">'
                          'Benjamin Schulz\'s work</a>',
                    parent_id=node.node_id, incoming_edge_config=google_edge)

    elif node.data_type == 'google.rlz.ap':
        # From https://source.chromium.org/chromium/chromium/src/+/master:rlz/lib/lib_values.cc
        rlz_aps = {
            # 'I7': 'IE_DEFAULT_SEARCH',
            'I7': 'IE Default search',
            # 'W1': 'IE_HOME_PAGE',
            'W1': 'IE Home Page',
            # 'T4': 'IETB_SEARCH_BOX',
            'T4': 'IE Toolbar Search Box',
            'Q1': 'QUICK_SEARCqH_BOX',
            'D1': 'GD_DESKBAND',
            'D2': 'GD_SEARCH_GADGET',
            'D3': 'GD_WEB_SERVER',
            'D4': 'GD_OUTLOOK',
            # 'C1': 'CHROME_OMNIBOX',
            'C1': 'Chrome Omnibox',
            # 'C2': 'CHROME_HOME_PAGE',
            'C2': 'Chrome Home Page',
            # 'B2': 'FFTB2_BOX',
            'B2': 'Firefox Toolbar v2',
            # 'B3': 'FFTB3_BOX',
            'B3': 'Firefox Toolbar v3',
            'N1': 'PINYIN_IME_BHO',
            'G1': 'IGOOGLE_WEBPAGE',
            'H1': 'MOBILE_IDLE_SCREEN_BLACKBERRY',
            'H2': 'MOBILE_IDLE_SCREEN_WINMOB',
            'H3': 'MOBILE_IDLE_SCREEN_SYMBIAN',
            # 'R0': 'FF_HOME_PAGE',
            'R0': 'Firefox Home Page',
            # 'R1': 'FF_SEARCH_BOX',
            'R1': 'Firefox Search box',
            'R2': 'IE_BROWSED_PAGE',
            'R3': 'QSB_WIN_BOX',
            'R4': 'WEBAPPS_CALENDAR',
            'R5': 'WEBAPPS_DOCS',
            'R6': 'WEBAPPS_GMAIL',
            'R7': 'IETB_LINKDOCTOR',
            'R8': 'FFTB_LINKDOCTOR',
            'T7': 'IETB7_SEARCH_BOX',
            'T8': 'TB8_SEARCH_BOX',
            'C3': 'CHROME_FRAME',
            'V1': 'PARTNER_AP_1',
            'V2': 'PARTNER_AP_2',
            'V3': 'PARTNER_AP_3',
            'V4': 'PARTNER_AP_4',
            'V5': 'PARTNER_AP_5',
            # 'C5': 'CHROME_MAC_OMNIBOX',
            'C5': 'Chrome Omnibox on Mac',
            # 'C6': 'CHROME_MAC_HOME_PAGE',
            'C6': 'Chrome home page on Mac',
            # 'CA': 'CHROMEOS_OMNIBOX',
            'CA': 'ChromeOS Omnibox',
            # 'CB': 'CHROMEOS_HOME_PAGE',
            'CB': 'ChromeOS Home Page',
            'CC': 'CHROMEOS_APP_LIST',
            # 'C9': 'CHROME_IOS_OMNIBOX_TABLET',
            'C9': 'Chrome Omnibox on iOS Tablet',
            # 'CD': 'CHROME_IOS_OMNIBOX_MOBILE',
            'CD': 'Chrome Omnibox on iOS Mobile',
            'C7': 'CHROME_APP_LIST',
            'C8': 'CHROME_MAC_APP_LIST',
            'RQ': 'UNDEFINED_AP_Q',
            'RR': 'UNDEFINED_AP_R',
            'RS': 'UNDEFINED_AP_S',
            'RT': 'UNDEFINED_AP_T',
            'RU': 'UNDEFINED_AP_U',
            'RV': 'UNDEFINED_AP_V',
            'RW': 'UNDEFINED_AP_W',
            'RX': 'UNDEFINED_AP_X',
            'RY': 'UNDEFINED_AP_Y',
            'RZ': 'UNDEFINED_AP_Z',
            'U0': 'PACK_AP0',
            'U1': 'PACK_AP1',
            'U2': 'PACK_AP2',
            'U3': 'PACK_AP3',
            'U4': 'PACK_AP4',
            'U5': 'PACK_AP5',
            'U6': 'PACK_AP6',
            'U7': 'PACK_AP7',
            'U8': 'PACK_AP8',
            'U9': 'PACK_AP9',
            'UA': 'PACK_AP10',
            'UB': 'PACK_AP11',
            'UC': 'PACK_AP12',
            'UD': 'PACK_AP13'
        }

        unfurl.add_to_queue(
            data_type='descriptor', key=node.value, value=rlz_aps.get(node.value, 'Unknown'),
            label=f'Search performed using {rlz_aps.get(node.value, "unknown application")}',
            parent_id=node.node_id, incoming_edge_config=google_edge)

    elif node.data_type == 'google.ei' and node.key == 1:
        ei0_node = unfurl.check_sibling_nodes(node, data_type='google.ei', key=0, return_node=True)

        if ei0_node:
            # The first ei value is epoch seconds, the second has the fractional (micro) seconds.
            # Concatenating them here as strings lets the timestamp parser convert it later.
            ei_timestamp = str(ei0_node.value) + str(node.value)

            unfurl.add_to_queue(
                data_type='epoch-microseconds', key=None, value=ei_timestamp, label=f'ei Timestamp: {ei_timestamp}',
                hover='The first two values combined in the <b>ei</b> parameter are thought to be the timestamp of '
                      'when the session began. The first (ei-0) contains the full seconds portion of the timestamp '
                      'and the second (ei-1) contains the fractional seconds.',
                parent_id=[node.node_id, ei0_node.node_id],
                incoming_edge_config=google_edge)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json

import logging
log = logging.getLogger(__name__)

json_edge = {
    'color': {
        'color': '#A7A7A7'
    },
    'title': 'JSON Parsing Functions',
    'label': 'JSON'
}


def run(unfurl, node):

    if node.data_type in ('url.query.pair', 'string'):
        try:
            json_obj = json.loads(node.value)
            assert isinstance(json_obj, dict), 'Loaded object should be a dict'

        except json.JSONDecodeError:
            return

        except AssertionError:
            # Likely a "valid" JSON of 23 or 1 or some single value; not what we want
            return

        try:

            for json_key, json_value in json_obj.items():
                unfurl.add_to_queue(
                    data_type='json', key=json_key, value=json_value, label=f'{json_key}: {json_value}',
                    hover='This was parsed as JavaScript Object Notation (JSON), <br>'
                          'which uses human-readable text to store and transmit data objects',
                    parent_id=node.node_id, incoming_edge_config=json_edge)

        except Exception as e:
            log.exception(f'Exception parsing JSON string: {e}')

    elif node.data_type == 'json':
        node_value = node.value
        if isinstance(node_value, str):
            try:
                node_value = json.loads(node_value)
            except json.JSONDecodeError:
                pass

        if isinstance(node_value, dict):
            try:
                for key, value in node_value.items():
                    unfurl.add_to_queue(
                        data_type='json', key=key, value=value, label=f'{key}: {value}',
                        hover='This was parsed as JavaScript Object Notation (JSON), <br>'
                              'which uses human-readable text to store and transmit data objects',
                        parent_id=node.node_id, incoming_edge_config=json_edge)

            except Exception as e:
                log.exception(f'Exception parsing JSON: {e}')



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import time


youtube_edge = {
    'color': {
        'color': '#ff0000'
    },
    'title': 'Youtube Parsing Functions',
    'label': 'YT'
}


def format_seconds(seconds):
    if seconds.endswith('s'):
        seconds = seconds[:-1]
    try:
        seconds = int(seconds)
    except ValueError:
        return None

    if seconds < 60:
        return f'{seconds} seconds'
    elif seconds < 3600:
        return time.strftime('%M:%S', time.gmtime(seconds)) + ' (mm:ss)'
    else:
        return time.strftime('%H:%M:%S', time.gmtime(seconds)) + ' (hh:mm:ss)'


def run(unfurl, node):
    youtube_domains = ['youtube.com', 'youtu.be']
    if any(youtube_domain in unfurl.find_preceding_domain(node) for youtube_domain in youtube_domains):
        if node.key == 't' or node.key == 'time_continue' or node.key == 'start':
            time_formatted = format_seconds(node.value)

            if time_formatted:
                param_text = f'Video will start playing at {time_formatted}'
                unfurl.add_to_queue(
                    data_type='descriptor', key=None, value=node.value, label=param_text,
                    parent_id=node.node_id, incoming_edge_config=youtube_edge)

        if node.key == 'v' or (node.data_type == 'url.path.segment' and len(node.value) == 11) or \
                (node.data_type == 'url.path' and len(node.value[1:].split('/')) == 1 and len(node.value) == 12):
            video_id = node.value
            if node.data_type == 'url.path':
                video_id = video_id[1:]
            param_text = f'Video ID: {video_id}'
            unfurl.add_to_queue(
                data_type='descriptor', key=None, value=node.value, label=param_text,
                parent_id=node.node_id, incoming_edge_config=youtube_edge)



"""
Python methods for importing and exporting '.proto' files from the BBP type
definition format.
"""

# TODO get custom exceptions for these methods

import io
import re
import logging
from unfurl.lib.blackboxprotobuf.lib.exceptions import TypedefException
from unfurl.lib.blackboxprotobuf.lib.api import sort_typedef

PROTO_FILE_TYPE_MAP = {
    "uint": "uint64",
    "int": "int64",
    "sint": "sint64",
    "fixed32": "fixed32",
    "sfixed32": "sfixed32",
    "float": "float",
    "fixed64": "fixed64",
    "sfixed64": "sfixed64",
    "double": "double",
    "bytes": "bytes",
    "bytes_hex": "bytes",
    "string": "string",
}

PACKABLE_TYPES = [
    "uint",
    "int",
    "sint",
    "fixed32",
    "sfixed32",
    "float",
    "fixed64",
    "sfixed64",
    "double",
]

# Inverse of the above, but we have to include more types
PROTO_FILE_TYPE_TO_BBP = {
    "double": "double",
    "float": "float",
    "int32": "int",
    "int64": "int",
    "uint32": "uint",
    "uint64": "uint",
    "sint32": "sint",
    "sint64": "sint",
    "fixed32": "fixed32",
    "fixed64": "fixed64",
    "sfixed32": "sfixed32",
    "sfixed64": "sfixed64",
    "bool": "uint",
    "string": "string",
    # should be default_binary_type, but can't handle that well here
    "bytes": "bytes",
}

NAME_REGEX = re.compile(r"\A[a-zA-Z_][a-zA-Z0-9_]*\Z")

# add packed types to the list
for packable_type in PACKABLE_TYPES:
    packed_type = "packed_" + packable_type
    PROTO_FILE_TYPE_MAP[packed_type] = PROTO_FILE_TYPE_MAP[packable_type]


def _print_message(message_name, typedef, output_file, depth=0):
    indent = u"  " * depth
    if not NAME_REGEX.match(message_name):
        raise TypedefException("Message name: %s is not valid" % message_name)

    # sort typedef for better looking output
    typedef = sort_typedef(typedef)

    message_name = message_name.strip()
    output_file.write(u"\n")
    output_file.write(indent)
    output_file.write(u"message %s {\n" % message_name)
    for field_number, field_typedef in typedef.items():
        # TODO Default to all fields as repeated? or optional
        proto_type = None
        field_name = None
        field_options = ""

        # a repeated field with one element is indistinduishable from a
        # repeated field so we just put repeated if we have proof that it is
        # repeatable, but this might be wrong sometimes
        # maybe some sort of protobuf discovery tool can detect this
        is_repeated = field_typedef.get("seen_repeated", False)

        if "name" in field_typedef and field_typedef["name"] != "":
            field_name = field_typedef["name"]
            field_name = field_name.strip()
            if not NAME_REGEX.match(field_name):
                field_name = None
        if field_name is None:
            field_name = u"field%s" % field_number

        if field_typedef["type"] == "message":
            # If we have multiple typedefs, this means is something like the Any
            # message, and has to be manually reparsed to each type
            if "alt_typedefs" in field_typedef:
                proto_type = "bytes"
            else:
                proto_type = field_name + "_type"
                _print_message(
                    proto_type, field_typedef["message_typedef"], output_file, depth + 1
                )
        else:
            if field_typedef["type"] not in PROTO_FILE_TYPE_MAP:
                raise TypedefException(
                    "Type %s does not have a mapping to protobuf types."
                    % field_typedef["type"]
                )
            proto_type = PROTO_FILE_TYPE_MAP[field_typedef["type"]]

        # we're using proto3 syntax. Repeated numeric fields are packed by default
        # if it's repeated and not packed, then make sure we specify it's not packed
        if is_repeated and field_typedef["type"] in PACKABLE_TYPES:
            field_options = u" [packed=false]"
        # if it's a packed type, we'll explicitoly set that too, can't hurt
        elif field_typedef["type"].startswith("packed_"):
            field_options = u" [packed=true]"
            is_repeated = True

        output_file.write(indent)
        output_file.write(
            u"  %s%s %s = %s%s;\n"
            % (
                "repeated " if is_repeated else "",
                proto_type,
                field_name,
                field_number,
                field_options,
            )
        )

    output_file.write(indent)
    output_file.write(u"}\n\n")


def export_proto(typedef_map, output_filename=None, output_file=None, package=None):
    """Export the given type definitons as a '.proto' file. Typedefs are
    expected as a dictionary of {'message_name': typedef }

    Write to output_file or output_filename if provided, otherwise return a string
    output_filename will be overwritten if it exists
    """
    return_string = False
    if output_filename is not None:
        output_file = io.open(output_filename, "w+")

    if output_file is None:
        return_string = True
        output_file = io.StringIO()

    # preamble
    output_file.write(u'syntax = "proto3";\n\n')
    if package:
        output_file.write(u"package %s;\n\n" % package)

    for typedef_name, typedef in typedef_map.items():
        _print_message(typedef_name, typedef, output_file)

    if return_string:
        return output_file.getvalue()
    # close the file if we opened it
    elif output_filename is not None:
        output_file.close()
    return None


MESSAGE_START_REGEX = re.compile(r"^message +([a-zA-Z_0-9]+) *{.*")
FIELD_REGEX = re.compile(
    r"^ *(repeated|optional|required)? *([a-zA-Z0-9_]+) +([a-zA-Z0-9_]+) += +([0-9]+) *(\[[a-z]+=[a-z]*\])?.*;.*$"
)
SYNTAX_REGEX = re.compile(r'^ *syntax += +"(proto\d)" *;.*')
ENUM_REGEX = re.compile(r"^ *enum +([a-zA-Z0-9_]+) *{.*")
PACKAGE_REGEX = re.compile(r"^ *package +([a-zA-Z0-9_.]+) *;.*")


def import_proto(config, input_string=None, input_filename=None, input_file=None):
    typedef_map = {}
    if input_string is not None:
        input_file = io.StringIO(input_string)
    if input_file is None and input_filename is not None:
        input_file = io.open(input_filename, "r")
    if input_file is None:
        raise ValueError("No file provided to import_proto")

    syntax_version = "proto2"
    package_prefix = ""
    enum_names = []
    message_trees = []
    message_names = []

    line = input_file.readline()
    while line:
        line = line.strip()

        if line.startswith("syntax") and SYNTAX_REGEX.match(line):
            syntax_version = SYNTAX_REGEX.match(line).group(1)

        elif line.startswith("package") and PACKAGE_REGEX.match(line):
            package_prefix = PACKAGE_REGEX.match(line).group(1) + "."

        elif line.startswith("import"):
            logging.warn(
                "Proto file has import which is not supported "
                "by the parser. Ensure the imported files are "
                "processed first: %s",
                line,
            )
        elif line.startswith("enum") and ENUM_REGEX.match(line):
            enum_name = _parse_enum(line, input_file)
            enum_names.append(enum_name)

        elif line.startswith("message") and MESSAGE_START_REGEX.match(line):
            message_tree = _preparse_message(line, input_file)
            message_trees.append(message_tree)

        line = input_file.readline()

    # TODO parse the message data
    for tree in message_trees:
        new_message_names, new_enum_names = _collect_names(package_prefix, tree)
        enum_names += new_enum_names
        message_names += new_message_names

    logging.debug("Got the following enum_names: %s", enum_names)
    logging.debug("Got the following message_names: %s", message_names)

    for tree in message_trees:
        _parse_message(
            tree,
            typedef_map,
            message_names,
            enum_names,
            package_prefix,
            syntax_version == "proto3",
            config,
        )

    return typedef_map


def _parse_enum(line, input_file):
    """Parse an enum out of the file. Goes from enum declaration to next }
    Returns the enum's name
    """
    enum_name = ENUM_REGEX.match(line).group(1)
    # parse until the next '}'
    while "}" not in line:
        line = input_file.readline()
        if not line:
            raise ValueError("Did not find close of enum")
    return enum_name


def _preparse_message(line, input_file):
    """Parse out a message name and the lines that make it up"""
    message_name = MESSAGE_START_REGEX.match(line).group(1)
    message_lines = []
    inner_enums = []
    inner_messages = []

    while "}" not in line:
        line = input_file.readline()
        if not line:
            raise ValueError("Did not find close of message")

        line = line.strip()
        if line.startswith("enum") and ENUM_REGEX.match(line):
            enum_name = _parse_enum(line, input_file)
            inner_enums.append(enum_name)

        elif line.startswith("message") and MESSAGE_START_REGEX.match(line):
            message_tree = _preparse_message(line, input_file)
            inner_messages.append(message_tree)
        # not an inner enum or message
        else:
            message_lines.append(line)

    return {
        "name": message_name,
        "data": message_lines,
        "enums": inner_enums,
        "inner_messages": inner_messages,
    }


def _collect_names(prefix, message_tree):
    message_names = []
    enum_names = []

    name = prefix + message_tree["name"]
    message_names.append(name)
    for enum_name in message_tree["enums"]:
        enum_names.append(prefix + enum_name)
    for inner_message in message_tree["inner_messages"]:
        new_message_names, new_enum_names = _collect_names(name + ".", inner_message)
        message_names += new_message_names
        enum_names += new_enum_names
    return message_names, enum_names


def _check_message_name(current_path, name, known_message_names, config):
    # Verify message name against preparsed message names and global
    # known_messages
    # For example, if we have:
    #   Message.InnerMesage
    # referenced from:
    #    PackageA.Message2
    # we would look up:
    # PackageA.Message2.Message.InnerMessage
    # PackageA.Message.InnerMessage
    # should also work for enums
    if name in config.known_types:
        return True
    # search for anything under a common prefix in known_message_names
    logging.debug("Testing message name: %s", name)

    prefix_options = [""]
    for part in current_path.split("."):
        if part:
            prefix_options = [prefix_options[0] + part + "."] + prefix_options

    logging.debug("prefix_options: %s", prefix_options)
    for prefix in prefix_options:
        logging.debug("Testing message name: %s", prefix + name)
        if prefix + name in known_message_names:
            return prefix + name
        # remove the last bit of the prefix
        if "." not in prefix:
            break
        prefix = ".".join(prefix.split(".")[:-1])
    logging.debug(
        "Message %s not found from %s Known names are: %s",
        name,
        current_path,
        known_message_names,
    )
    return None


def _parse_message(
    message_tree, typdef_map, known_message_names, enum_names, prefix, is_proto3, config
):
    message_typedef = {}
    message_name = prefix + message_tree["name"]
    prefix = message_name + "."
    # parse the actual message fields
    for line in message_tree["data"]:
        # lines should already be stripped and should not have messages or enums
        # logging.debug("Line before assert: %s", line)
        assert all([not line.strip().startswith(x) for x in ["message ", "enum "]])
        # Check if the line matches the field regex
        match = FIELD_REGEX.match(line)
        if match:
            field_number, field_typedef = _parse_field(
                match, known_message_names, enum_names, prefix, is_proto3, config
            )
            message_typedef[field_number] = field_typedef

    # add the messsage to tyep returned typedefs
    logging.debug("Adding message %s to typedef maps", message_name)
    typdef_map[message_name] = message_typedef

    for inner_message in message_tree["inner_messages"]:
        # TODO prefix should be added to?
        _parse_message(
            inner_message,
            typdef_map,
            known_message_names,
            enum_names,
            prefix,
            is_proto3,
            config,
        )


# parse a field into a dictionary for the typedef
def _parse_field(match, known_message_names, enum_names, prefix, is_proto3, config):
    typedef = {}

    field_name = match.group(3)
    if not field_name:
        raise ValueError("Could not parse field name from line: %s" % match)
    typedef["name"] = field_name
    field_number = match.group(4)
    if not field_number:
        raise ValueError("Could not parse field number from line: %s" % match)

    # figure out repeated
    field_rule = match.group(1)
    is_repeated = False
    if field_rule and "repeated" in field_rule:
        is_repeated = True
        typedef["seen_repeated"] = True

    field_type = match.group(2)
    if not field_type:
        raise ValueError("Could not parse field type from line: %s" % match)
    # check normal types
    bbp_type = PROTO_FILE_TYPE_TO_BBP.get(field_type, None)
    if not bbp_type:
        logging.debug("Got non-basic type: %s, checking enums", field_type)
        # check enum names
        if _check_message_name(prefix, field_type, enum_names, config):
            # enum = uint
            bbp_type = "uint"
    if not bbp_type:
        # Not enum or normal type, check messages
        message_name = _check_message_name(
            prefix, field_type, known_message_names, config
        )
        if message_name:
            bbp_type = "message"
            typedef["message_type_name"] = message_name

    if not bbp_type:
        # If we don't have a type now, then fail
        raise ValueError(
            "Could not get a type for field %s: %s" % (field_name, field_type)
        )

    # figure out packed
    # default based on repeated + proto3, fallback to options
    field_options = match.group(5)
    is_packed = is_repeated and is_proto3 and (field_type in PACKABLE_TYPES)
    if is_packed and field_options and "packed=false" in field_options:
        is_packed = False
    elif is_repeated and field_options and "packed=true" in field_options:
        is_packed = True

    # make sure the type lines up with packable
    if is_packed and bbp_type not in PACKABLE_TYPES:
        raise ValueError(
            "Field %s set as packable, but not a packable type: %s"
            % (field_name, bbp_type)
        )
    if is_packed:
        bbp_type = "packed_" + bbp_type

    typedef["type"] = bbp_type

    logging.debug("Parsed field number %s: %s", field_number, typedef)
    return field_number, typedef



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import struct
from unfurl import utils

metasploit_edge = {
    'color': {
        'color': 'red'
    },
    'title': 'Metasploit Parsing Functions',
    'label': 'M'
}

# Reference: https://github.com/rapid7/metasploit-framework/blob/master/lib/rex/payloads/meterpreter/uri_checksum.rb
# Metasploit 8-bit checksums
#  URI_CHECKSUM_INITW = 92  # Windows
#  URI_CHECKSUM_INITN = 92  # Native (same as Windows)
#  URI_CHECKSUM_INITP = 80  # Python
#  URI_CHECKSUM_INITJ = 88  # Java
#  URI_CHECKSUM_CONN = 98  # Existing session
#  URI_CHECKSUM_INIT_CONN = 95  # New stageless session

metasploit_checksums = {
    80: 'Python',
    88: 'Java',
    92: 'Windows',
    95: 'New stageless session',
    98: 'Existing session'
}

# Reference: https://github.com/rapid7/metasploit-framework/blob/master/lib/msf/core/payload/uuid.rb
architectures = {
    # 0: 'nil',
    1: 'ARCH_X86',
    2: 'ARCH_X64',
    3: 'ARCH_X64',
    4: 'ARCH_MIPS',
    5: 'ARCH_MIPSLE',
    6: 'ARCH_MIPSBE',
    7: 'ARCH_PPC',
    8: 'ARCH_PPC64',
    9: 'ARCH_CBEA',
    10: 'ARCH_CBEA64',
    11: 'ARCH_SPARC',
    12: 'ARCH_ARMLE',
    13: 'ARCH_ARMBE',
    14: 'ARCH_CMD',
    15: 'ARCH_PHP',
    16: 'ARCH_TTY',
    17: 'ARCH_JAVA',
    18: 'ARCH_RUBY',
    19: 'ARCH_DALVIK',
    20: 'ARCH_PYTHON',
    21: 'ARCH_NODEJS',
    22: 'ARCH_FIREFOX',
    23: 'ARCH_ZARCH',
    24: 'ARCH_AARCH64',
    25: 'ARCH_MIPS64',
    26: 'ARCH_PPC64LE',
    27: 'ARCH_R',
    28: 'ARCH_PPCE500V2'
}

# Reference: https://github.com/rapid7/metasploit-framework/blob/master/lib/msf/core/payload/uuid.rb
platforms = {
    0: 'nil',
    1: 'windows',
    2: 'netware',
    3: 'android',
    4: 'java',
    5: 'ruby',
    6: 'linux',
    7: 'cisco',
    8: 'solaris',
    9: 'osx',
    10: 'bsd',
    11: 'openbsd',
    12: 'bsdi',
    13: 'netbsd',
    14: 'freebsd',
    15: 'aix',
    16: 'hpux',
    17: 'irix',
    18: 'unix',
    19: 'php',
    20: 'js',
    21: 'python',
    22: 'nodejs',
    23: 'firefox',
    24: 'r',
    25: 'apple_ios',
    26: 'juniper',
    27: 'unifi',
    28: 'brocade',
    29: 'mikrotik',
    30: 'arista'
}


def run(unfurl, node):
    # Decoding a Metasploit Shellcode URL
    # Code is based off Didier's Stevens work:
    #  - Finding Metasploit & Cobalt Strike URLs
    #    (https://isc.sans.edu/forums/diary/Finding+Metasploit+Cobalt+Strike+URLs/27204/)
    #  - metatool.py (https://github.com/DidierStevens/Beta/blob/master/metatool.py#L219)
    # Metasploit code:
    #  - https://github.com/rapid7/metasploit-framework/blob/master/lib/rex/payloads/meterpreter/uri_checksum.rb
    if node.data_type == 'url.path' and len(node.value) == 5:
        checksum = sum(map(ord, node.value[1:])) % 0x100
        metasploit_checksum = metasploit_checksums.get(checksum)

        if not metasploit_checksum:
            return

        unfurl.add_to_queue(
            data_type='descriptor', key=None, value=None,
            label=f'Matches Metasploit URL checksum for {metasploit_checksum}',
            hover='Shellcode URLs generated by Metasploit and Cobalt Strike have a path <br>'
                  'of four characters and are validated by a checksum. This URL path fits <br>'
                  f'the pattern and the checksum code matches <b>{metasploit_checksum}</b>. However, <br>'
                  'the checksum is short (8-bits) and <b>this could be a false positive</b>. '
                  '[<a href="https://isc.sans.edu/forums/diary/Finding+Metasploit+Cobalt+Strike+URLs/27204/">ref</a>]',
            parent_id=node.node_id, incoming_edge_config=metasploit_edge)

    # Decoding a Metasploit Payload UUID
    # Ex: http://127.0.0.1/aAF9cpWMQPb-3f_cq1FoJA040uMw26kAnvroJdztpVzDr... (path of at least 22 chars)
    # Code is based off Didier's Stevens work:
    #  - Metasploit's Payload UUID (https://isc.sans.edu/forums/diary/Metasploits+Payload+UUID/23555/)
    #  - metatool.py (https://github.com/DidierStevens/Beta/blob/master/metatool.py#L254)
    # Metasploit docs/code:
    #  - Payload UUID by HD Moore (https://github.com/rapid7/metasploit-framework/wiki/Payload-UUID)
    #  - uuid.rb (https://github.com/rapid7/metasploit-framework/blob/master/lib/msf/core/payload/uuid.rb)

    if node.data_type == 'url.path' and len(node.value) >= 23:

        cleaned_path = node.value.lstrip('/').rstrip('/')

        # URL path should be base64 encoded
        if not (utils.digits_re.search(cleaned_path) and utils.letters_re.search(cleaned_path)
                and utils.urlsafe_b64_re.fullmatch(cleaned_path)):
            return

        decoded = base64.urlsafe_b64decode(cleaned_path[0:22] + '==')
        encoded_payload_uuid_format = '8sBBBBBBBB'
        puid, xor1, xor2, platform_xored, architecture_xored, ts1_xored, ts2_xored, ts3_xored, ts4_xored = \
            struct.unpack(encoded_payload_uuid_format, decoded)

        platform = platform_xored ^ xor1
        platform_name = platforms.get(platform, None)
        if platform_name is None:
            return

        architecture = architecture_xored ^ xor2
        architecture_name = architectures.get(architecture, None)
        if architecture_name is None:
            return

        timestamp = struct.unpack(
            '>I', bytes([ts1_xored ^ xor1, ts2_xored ^ xor2, ts3_xored ^ xor1, ts4_xored ^ xor2]))[0]

        # This is the lowest valid timestamp, per Metasploit code. Someone could customize it to be different, but
        # filtering on this seems like a good way to reduce false positives.
        # Source: https://github.com/rapid7/metasploit-framework/blob/master/lib/msf/core/payload/uuid.rb#L90
        if timestamp < 1420070400:
            return

        unfurl.add_to_queue(
            data_type='descriptor', key='Unique ID', value=bytes.hex(puid),
            hover='Metasploit\'s <b>puid</b> is an optional 8-byte string that<br> '
                  'can be used as the unique payload ID. '
                  '[<a href="https://isc.sans.edu/forums/diary/Metasploits+Payload+UUID/23555/">ref</a> & '
                  '<a href="https://github.com/rapid7/metasploit-framework/blob/master/lib/msf/core/payload/uuid.rb">'
                  'ref</a>]',
            parent_id=node.node_id, incoming_edge_config=metasploit_edge)

        unfurl.add_to_queue(
            data_type='descriptor', key='Platform', value=platform_name.capitalize(),
            hover='The operating system platform for this Metasploit payload.',
            parent_id=node.node_id, incoming_edge_config=metasploit_edge)

        unfurl.add_to_queue(
            data_type='descriptor', key='Architecture', value=architecture_name[5:],
            hover='The hardware architecture for this Metasploit payload.',
            parent_id=node.node_id, incoming_edge_config=metasploit_edge)

        unfurl.add_to_queue(
            data_type='epoch-seconds', key='Timestamp', value=timestamp,
            hover='The timestamp (in Epoch seconds) the payload was generated.',
            parent_id=node.node_id, incoming_edge_config=metasploit_edge)



"""Methods for easy encoding and decoding of messages"""

import re
import six
import json
import collections
import unfurl.lib.blackboxprotobuf.lib.types.length_delim
import unfurl.lib.blackboxprotobuf.lib.config
from unfurl.lib.blackboxprotobuf.lib.exceptions import TypedefException


def decode_message(buf, message_type=None, config=None):
    """Decode a message to a Python dictionary.
    Returns tuple of (values, types)
    """

    if config is None:
        config = unfurl.lib.blackboxprotobuf.lib.config.default

    if isinstance(buf, bytearray):
        buf = bytes(buf)
    buf = six.ensure_binary(buf)
    if message_type is None or isinstance(message_type, str):
        if message_type not in config.known_types:
            message_type = {}
        else:
            message_type = config.known_types[message_type]

    value, typedef, _ = unfurl.lib.blackboxprotobuf.lib.types.length_delim.decode_message(
        buf, config, message_type
    )
    return value, typedef


# TODO add explicit validation of values to message type
def encode_message(value, message_type, config=None):
    """Encodes a python dictionary to a message.
    Returns a bytearray
    """

    if config is None:
        config = unfurl.lib.blackboxprotobuf.lib.config.default
    return bytes(
        unfurl.lib.blackboxprotobuf.lib.types.length_delim.encode_message(
            value, config, message_type
        )
    )


def protobuf_to_json(*args, **kwargs):
    """Encode to python dictionary and dump to JSON.
    Takes same arguments as decode_message
    """
    value, message_type = decode_message(*args, **kwargs)
    value = json_safe_transform(
        value, message_type, False, config=kwargs.get("config", None)
    )
    value = sort_output(value, message_type, config=kwargs.get("config", None))
    _annotate_typedef(message_type, value)
    message_type = sort_typedef(message_type)
    return json.dumps(value, indent=2), message_type


def protobuf_from_json(json_str, message_type, *args, **kwargs):
    """Decode JSON string to JSON and then to protobuf.
    Takes same arguments as encode_message
    """
    value = json.loads(json_str)
    _strip_typedef_annotations(message_type)
    value = json_safe_transform(value, message_type, True)
    return encode_message(value, message_type, *args, **kwargs)


def export_protofile(message_types, output_filename):
    """Export the give messages as ".proto" file.
    Expects a dictionary with {message_name: typedef} and a filename to
    write to.
    """
    unfurl.lib.blackboxprotobuf.lib.protofile.export_proto(
        message_types, output_filename=output_filename
    )


def import_protofile(input_filename, save_to_known=True, config=None):
    """import ".proto" files and returns the typedef map
    Expects a filename for the ".proto" file
    """
    if config is None:
        config = unfurl.lib.blackboxprotobuf.lib.config.default

    new_typedefs = unfurl.lib.blackboxprotobuf.lib.protofile.import_proto(
        config, input_filename=input_filename
    )
    if save_to_known:
        config.known_types.update(new_typedefs)
    else:
        return new_typedefs


NAME_REGEX = re.compile(r"\A[a-zA-Z][a-zA-Z0-9_]*\Z")


def validate_typedef(typedef, old_typedef=None, path=None, config=None):
    """Validate the typedef format. Optionally validate wiretype of a field
    number has not been changed
    """
    if path is None:
        path = []
    if config is None:
        config = unfurl.lib.blackboxprotobuf.lib.config.default

    int_keys = set()
    field_names = set()
    for field_number, field_typedef in typedef.items():
        alt_field_number = None
        if "-" in str(field_number):
            field_number, alt_field_number = field_number.split("-")

        # Validate field_number is a number
        if not str(field_number).isdigit():
            raise TypedefException("Field number must be a digit: %s" % field_number)
        field_number = str(field_number)

        field_path = path[:]
        field_path.append(field_number)

        # Check for duplicate field numbers
        if field_number in int_keys:
            raise TypedefException(
                "Duplicate field number: %s" % field_number, field_path
            )
        int_keys.add(field_number)

        # Must have a type field
        if "type" not in field_typedef:
            raise TypedefException(
                "Field number must have a type value: %s" % field_number, field_path
            )
        if alt_field_number is not None:
            if field_typedef["type"] != "message":
                raise TypedefException(
                    "Alt field number (%s) specified for non-message field: %s"
                    % (alt_field_number, field_number),
                    field_path,
                )

        valid_type_fields = [
            "type",
            "name",
            "message_typedef",
            "message_type_name",
            "alt_typedefs",
            "example_value_ignored",
            "seen_repeated",
        ]
        for key, value in field_typedef.items():
            # Check field keys against valid values
            if key not in valid_type_fields:
                raise TypedefException(
                    'Invalid field key "%s" for field number %s' % (key, field_number),
                    field_path,
                )
            if (
                key in ["message_typedef", "message_type_name"]
                and not field_typedef["type"] == "message"
            ):
                raise TypedefException(
                    'Invalid field key "%s" for field number %s' % (key, field_number),
                    field_path,
                )
            if key == "group_typedef" and not field_typedef["type"] == "group":
                raise TypedefException(
                    'Invalid field key "%s" for field number %s' % (key, field_number),
                    field_path,
                )

            # Validate type value
            if key == "type":
                if value not in unfurl.lib.blackboxprotobuf.lib.types.type_maps.WIRETYPES:
                    raise TypedefException(
                        'Invalid type "%s" for field number %s' % (value, field_number),
                        field_path,
                    )
            # Check for duplicate names
            if key == "name":
                if value.lower() in field_names:
                    raise TypedefException(
                        ('Duplicate field name "%s" for field ' "number %s")
                        % (value, field_number),
                        field_path,
                    )
                if not NAME_REGEX.match(value):
                    raise TypedefException(
                        (
                            'Invalid field name "%s" for field '
                            "number %s. Should match %s"
                        )
                        % (value, field_number, "[a-zA-Z_][a-zA-Z0-9_]*"),
                        field_path,
                    )
                if value != "":
                    field_names.add(value.lower())

            # Check if message type name is known
            if key == "message_type_name":
                if value not in config.known_types:
                    raise TypedefException(
                        (
                            'Message type "%s" for field number'
                            " %s is not known. Known types: %s"
                        )
                        % (value, field_number, config.known_types.keys()),
                        field_path,
                    )

            # Recursively validate inner typedefs
            if key in ["message_typedef", "group_typedef"]:
                if old_typedef is None:
                    validate_typedef(value, path=field_path, config=config)
                else:
                    validate_typedef(
                        value, old_typedef[field_number][key], path=field_path
                    )
            if key == "alt_typedefs":
                for alt_field_number, alt_typedef in value.items():
                    # TODO validate alt_typedefs against old typedefs?
                    validate_typedef(alt_typedef, path=field_path, config=config)

    if old_typedef is not None:
        wiretype_map = {}
        for field_number, value in old_typedef.items():
            wiretype_map[
                int(field_number)
            ] = unfurl.lib.blackboxprotobuf.lib.types.type_maps.WIRETYPES[value["type"]]
        for field_number, value in typedef.items():
            field_path = path[:]
            field_path.append(str(field_number))
            if int(field_number) in wiretype_map:
                old_wiretype = wiretype_map[int(field_number)]
                if (
                    old_wiretype
                    != unfurl.lib.blackboxprotobuf.lib.types.type_maps.WIRETYPES[value["type"]]
                ):
                    raise TypedefException(
                        (
                            "Wiretype for field number %s does"
                            " not match old type definition"
                        )
                        % field_number,
                        field_path,
                    )


def json_safe_transform(values, typedef, toBytes, config=None):
    """JSON doesn't handle bytes type well. We want to go through and encode
    every bytes type as latin1 to get a semi readable text"""

    if config is None:
        config = unfurl.lib.blackboxprotobuf.lib.config.default
    name_map = {
        item["name"]: number
        for number, item in typedef.items()
        if item.get("name", None)
    }
    for name, value in values.items():
        alt_number = None
        base_name = name
        if "-" in name:
            base_name, alt_number = name.split("-")

        if base_name in name_map:
            field_number = name_map[base_name]
        else:
            field_number = base_name

        field_type = typedef[field_number]["type"]
        if field_type == "message":
            field_typedef = _get_typedef_for_message(typedef[field_number], config)
            if alt_number is not None:
                # if we have an alt type, then let's look that up instead
                if alt_number not in typedef[field_number].get("alt_typedefs", {}):
                    raise TypedefException(
                        (
                            "Provided alt field name/number "
                            "%s is not valid for field_number %s"
                        )
                        % (alt_number, field_number)
                    )
                field_type = typedef[field_number]["alt_typedefs"][alt_number]
                if isinstance(field_type, dict):
                    field_typedef = field_type
                    field_type = "message"

        is_list = isinstance(value, list)
        field_values = value if is_list else [value]
        for i, field_value in enumerate(field_values):
            if field_type == "bytes":
                if toBytes:
                    field_values[i] = field_value.encode("latin1")
                else:
                    field_values[i] = field_value.decode("latin1")
            elif field_type == "message":
                field_values[i] = json_safe_transform(
                    field_value,
                    field_typedef,
                    toBytes,
                    config=config,
                )

        # convert back to single value if needed
        if not is_list:
            values[name] = field_values[0]
        else:
            values[name] = field_values
    return values


def _get_typedef_for_message(field_typedef, config):
    assert field_typedef["type"] == "message"
    if "message_typedef" in field_typedef:
        return field_typedef["message_typedef"]
    elif field_typedef.get("message_type_name"):
        if field_typedef["message_type_name"] not in config.known_types:
            raise TypedefException(
                "Got 'message_type_name' not in known_messages: %s"
                % field_typedef["message_type_name"]
            )
        return config.known_types[field_typedef["message_type_name"]]
    else:
        raise TypedefException(
            "Got 'message' type without typedef or type name: %s" % field_typedef
        )


def sort_output(value, typedef, config=None):
    """Sort output by the field number in the typedef. Helps with readability
    in a JSON dump"""
    output_dict = collections.OrderedDict()
    if config is None:
        config = unfurl.lib.blackboxprotobuf.lib.config.default

    # Make a list of all the field names we have, aggregate together the alt fields as well
    field_names = {}
    for field_name in value.keys():
        if "-" in field_name:
            field_name_base, alt_number = field_name.split("-")
        else:
            field_name_base = field_name
            alt_number = None
        field_names.setdefault(field_name_base, []).append((field_name, alt_number))

    for field_number, field_def in sorted(typedef.items(), key=lambda t: int(t[0])):
        field_number = str(field_number)
        seen_field_names = field_names.get(field_number, [])

        # Try getting matching fields by name as well
        if field_def.get("name", "") != "":
            field_name = field_def["name"]
            seen_field_names.extend(field_names.get(field_name, []))

        for field_name, alt_number in seen_field_names:
            field_type = field_def["type"]
            field_message_typedef = None
            if field_type == "message":
                field_message_typedef = _get_typedef_for_message(field_def, config)

            if alt_number is not None:
                if alt_number not in field_def["alt_typedefs"]:
                    raise TypedefException(
                        (
                            "Provided alt field name/number "
                            "%s is not valid for field_number %s"
                        )
                        % (alt_number, field_number)
                    )
                field_type = field_def["alt_typedefs"][alt_number]
                if isinstance(field_type, dict):
                    field_message_typedef = field_type
                    field_type = "message"

            if field_type == "message":
                if not isinstance(value[field_name], list):
                    output_dict[field_name] = sort_output(
                        value[field_name], field_message_typedef
                    )
                else:
                    output_dict[field_name] = []
                    for field_value in value[field_name]:
                        output_dict[field_name].append(
                            sort_output(field_value, field_message_typedef)
                        )
            else:
                output_dict[field_name] = value[field_name]

    return output_dict


def sort_typedef(typedef):
    """Sort output by field number and sub_keys so name then type is first"""

    TYPEDEF_KEY_ORDER = ["name", "type", "message_type_name", "example_value_ignored"]
    output_dict = collections.OrderedDict()

    for field_number, field_def in sorted(typedef.items(), key=lambda t: int(t[0])):
        output_field_def = collections.OrderedDict()
        field_def = field_def.copy()
        for key in TYPEDEF_KEY_ORDER:
            if key in field_def:
                output_field_def[key] = field_def[key]
                del field_def[key]
        for key, value in field_def.items():

            # TODO handle alt typedefs
            if key == "message_typedef":
                output_field_def[key] = sort_typedef(value)
            else:
                output_field_def[key] = value
        output_dict[field_number] = output_field_def
    return output_dict


def _annotate_typedef(typedef, message):
    """Add values from message into the typedef so it's easier to figure out
    which field when you're editing manually"""

    for field_number, field_def in typedef.items():
        field_value = None
        field_name = str(field_number)
        if field_name not in message and field_def.get("name", "") != "":
            field_name = field_def["name"]

        if field_name in message:
            field_value = message[field_name]

            # TODO handle alt typedefs
            if field_def["type"] == "message":
                if isinstance(field_value, list):
                    for value in field_value:
                        _annotate_typedef(field_def["message_typedef"], value)
                else:
                    _annotate_typedef(field_def["message_typedef"], field_value)
            else:
                field_def["example_value_ignored"] = field_value


def _strip_typedef_annotations(typedef):
    """Remove example values placed by _annotate_typedef"""
    for _, field_def in typedef.items():
        if "example_value_ignored" in field_def:
            del field_def["example_value_ignored"]
        if "message_typedef" in field_def:
            _strip_typedef_annotations(field_def["message_typedef"])



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mimetypes
import pycountry
import re
import urllib.parse

urlparse_edge = {
    'color': {
        'color': '#4d4d4d'
    },
    'title': 'URL Parsing Functions',
    'label': 'u'
}


def run(unfurl, node):

    if node.data_type == 'url':
        parsed_url = urllib.parse.urlparse(node.value)

        if parsed_url.netloc:
            if parsed_url.scheme:
                unfurl.add_to_queue(
                    data_type='url.scheme', key='Scheme', value=parsed_url.scheme,
                    hover='This is the URL <b>scheme</b>, per <a href="'
                          'https://tools.ietf.org/html/rfc3986" target="_blank">RFC3986</a>',
                    parent_id=node.node_id, incoming_edge_config=urlparse_edge)

            if parsed_url.netloc == parsed_url.hostname:
                unfurl.add_to_queue(
                    data_type='url.hostname', key=None, value=parsed_url.hostname,
                    hover='This is the <b>host</b> subcomponent of authority (also often called '
                          'netloc), per <a href="https://tools.ietf.org/html/rfc3986" target="_blank">RFC3986</a>',
                    parent_id=node.node_id, incoming_edge_config=urlparse_edge)
            else:
                unfurl.add_to_queue(
                    data_type='url.authority', key=None, value=parsed_url.netloc,
                    hover='This is the <b>authority</b> (also often called '
                          'netloc), per <a href="https://tools.ietf.org/html/rfc3986" target="_blank">RFC3986</a>',
                    parent_id=node.node_id, incoming_edge_config=urlparse_edge)

            if parsed_url.path and parsed_url.path != '/':
                unfurl.add_to_queue(
                    data_type='url.path', key=None, value=parsed_url.path,
                    hover='This is the URL <b>path</b>, per <a href="'
                          'https://tools.ietf.org/html/rfc3986" target="_blank">RFC3986</a>',
                    parent_id=node.node_id, incoming_edge_config=urlparse_edge)

            if parsed_url.params:
                unfurl.add_to_queue(
                    data_type='url.params', key=None, value=parsed_url.params,
                    hover='This is the URL path <b>parameters</b>, per <a href="'
                          'https://tools.ietf.org/html/rfc3986" target="_blank">RFC3986</a>',
                    parent_id=node.node_id, incoming_edge_config=urlparse_edge)

            if parsed_url.query:
                unfurl.add_to_queue(
                    data_type='url.query', key=None, value=parsed_url.query,
                    hover='This is the URL <b>query</b>, per <a href="'
                          'https://tools.ietf.org/html/rfc3986" target="_blank">RFC3986</a>',
                    parent_id=node.node_id, incoming_edge_config=urlparse_edge)

            if parsed_url.fragment:
                unfurl.add_to_queue(
                    data_type='url.fragment', key=None, value=parsed_url.fragment,
                    hover='This is the URL <b>fragment</b>, per <a href="'
                          'https://tools.ietf.org/html/rfc3986" target="_blank">RFC3986</a>',
                    parent_id=node.node_id, incoming_edge_config=urlparse_edge)

    elif node.data_type == 'url.path':
        path_segments = node.value.rstrip('/').split('/')
        if len(path_segments) > 2:
            for segment_number, path_segment in enumerate(path_segments):
                if path_segment != '':
                    unfurl.add_to_queue(
                        data_type='url.path.segment', key=segment_number, value=path_segment,
                        hover='This is a URL <b>path segment</b> (the URL path is split on "/"s). '
                              'Numbering starts at 1.', parent_id=node.node_id, incoming_edge_config=urlparse_edge)

    elif node.data_type == 'url.query' or node.data_type == 'url.fragment':
        parsed_qs = urllib.parse.parse_qs(node.value)
        for key, value in parsed_qs.items():
            assert type(value) is list, 'parsed_qs should result in type list, but did not.'
            # In the majority of cases, query string keys are unique, but the spec is ambiguous. In the case of
            # duplicate keys, urllib.parse.parsed_qs adds them to a list. Unfurl will loop over and create a
            # node for each value in that list of values (this is typically only one value, but could be more).
            for v in value:
                unfurl.add_to_queue(
                    data_type='url.query.pair', key=key, value=v, label=f'{key}: {v}',
                    parent_id=node.node_id, incoming_edge_config=urlparse_edge)

    elif node.data_type == 'url.params':
        split_params_re = re.compile(r'^(?P<key>[^=]+?)=(?P<value>[^=?]+)(?P<delim>[;,|])')
        split_params = split_params_re.match(node.value)
        if split_params:
            x = split_params.group('delim')
            parsed_params = node.value.split(x)
            for parsed_param in parsed_params:
                key, value = parsed_param.split('=')
                unfurl.add_to_queue(
                    data_type='url.param.pair', key=key, value=value,
                    parent_id=node.node_id, incoming_edge_config=urlparse_edge)

    # This should only occur when a URL node was parsed previously and netloc != hostname, which means there are
    # additional subcomponents of the authority to parse: user, password, and/or port. Reparsing this way lets the
    # simple case of authority = hostname look uncluttered, but still support all the other subcomponents if given.
    elif node.data_type == 'url.authority':
        # We need to add in a fake scheme here, as we stripped in the previous run and urlparse needs one.
        parsed_authority = urllib.parse.urlparse(f'https://{node.value}')

        if parsed_authority.username:
            unfurl.add_to_queue(
                data_type='url.username', key='Username', value=parsed_authority.username,
                hover='This is the <b>username</b> subcomponent of authority, '
                      'per <a href="https://tools.ietf.org/html/rfc3986" target="_blank">RFC3986</a>',
                parent_id=node.node_id, incoming_edge_config=urlparse_edge)

        if parsed_authority.password:
            unfurl.add_to_queue(
                data_type='url.password', key='Password', value=parsed_authority.password,
                hover='This is the <b>password</b> subcomponent of authority, '
                      'per <a href="https://tools.ietf.org/html/rfc3986" '
                      'target="_blank">RFC3986</a>',
                parent_id=node.node_id, incoming_edge_config=urlparse_edge)

        unfurl.add_to_queue(
            data_type='url.hostname', key='Host', value=parsed_authority.hostname,
            hover='This is the <b>host</b> subcomponent of authority (also often called '
                  'netloc), per <a href="https://tools.ietf.org/html/rfc3986" target="_blank">RFC3986</a>',
            parent_id=node.node_id, incoming_edge_config=urlparse_edge)

        if parsed_authority.port:
            unfurl.add_to_queue(
                data_type='url.port', key='Port', value=parsed_authority.port,
                hover='This is the <b>port</b> subcomponent of authority, '
                      'per <a href="https://tools.ietf.org/html/rfc3986" target="_blank">RFC3986</a>',
                parent_id=node.node_id, incoming_edge_config=urlparse_edge)

    elif node.data_type == 'url.query.pair' and node.key in ['l', 'lang', 'language', 'set-lang']:

        if len(node.value) == 2:
            language = pycountry.languages.get(alpha_2=node.value)
        elif len(node.value) == 3:
            language = pycountry.languages.get(alpha_3=node.value)
        else:
            return

        if language:
            unfurl.add_to_queue(
                data_type='descriptor', key='Language', value=language.name,
                hover='This is a generic parser based on common query-string patterns across websites',
                parent_id=node.node_id, incoming_edge_config=urlparse_edge)

    elif node.data_type == 'url.query.pair' and node.key in ['c', 'cc', 'country', 'country_code']:

        if len(node.value) == 2:
            country = pycountry.countries.get(alpha_2=node.value)
        elif len(node.value) == 3:
            country = pycountry.countries.get(alpha_3=node.value)
        else:
            return

        if country:
            unfurl.add_to_queue(
                data_type='descriptor', key='Country', value=country.name,
                hover='This is a generic parser based on common query-string patterns across websites',
                parent_id=node.node_id, incoming_edge_config=urlparse_edge)

    elif node.data_type == 'url.path.segment':
        for file_type in mimetypes.types_map.keys():
            if node.value.endswith(file_type):
                unfurl.add_to_queue(
                    data_type='file.name', key='File Name',
                    value=urllib.parse.unquote_plus(node.value[:-len(file_type)]),
                    parent_id=node.node_id, incoming_edge_config=urlparse_edge)
                unfurl.add_to_queue(
                    data_type='file.ext', key='File Extension', value=node.value[-len(file_type):],
                    hover=f'The data type typically associated with <b>{file_type}</b> '
                          f'is <b>{mimetypes.types_map[file_type]}</b>',
                    parent_id=node.node_id, incoming_edge_config=urlparse_edge)

    else:
        if not isinstance(node.value, str):
            return

        try:
            # If we can recognize another URL inside a value, parse it
            parsed_url = urllib.parse.urlparse(node.value)
            if (parsed_url.netloc and parsed_url.path) or (parsed_url.scheme and parsed_url.netloc):
                unfurl.add_to_queue(
                    data_type='url', key=None, value=node.value, parent_id=node.node_id,
                    incoming_edge_config=urlparse_edge)
                return
        except:
            # Guess it wasn't a URL
            pass

        # If the value contains more pairs of the form "a=b|c=d|e=f"
        pipe_delimited_pairs_re = re.compile(
            r'((?P<key>[^|=]+)=(?P<value>[^|=]+)\|)+(?P<last_key>[^|=]+)=(?P<last_value>[^|=]+)')
        m = pipe_delimited_pairs_re.fullmatch(node.value)
        if m:
            pipe_pairs = node.value.split('|')
            for pair in pipe_pairs:
                key, value = pair.split('=')
                unfurl.add_to_queue(
                    data_type='url.query.pair', key=key, value=value,
                    parent_id=node.node_id, incoming_edge_config=urlparse_edge)
            return

        # If the value contains more values in the form "a|b|c|d|e|f"
        pipe_delimited_values_re = re.compile(
            r'((?P<value>[^|]+)\|)+(?P<last_value>[^|]+)')
        m = pipe_delimited_values_re.fullmatch(node.value)
        if m:
            pipe_values = node.value.split('|')
            for value in pipe_values:
                unfurl.add_to_queue(
                    data_type='string', key=None, value=value,
                    parent_id=node.node_id, incoming_edge_config=urlparse_edge)
            return

        # If the value contains more pairs of the form "a=b&c=d&e=f"
        amp_delimited_pairs_re = re.compile(
            r'((?P<key>[^&=]+)=(?P<value>[^&=]+)&)+(?P<last_key>[^&=]+)=(?P<last_value>[^&=]+)')
        m = amp_delimited_pairs_re.fullmatch(node.value)
        if m:
            amp_pairs = node.value.split('&')
            for pair in amp_pairs:
                key, value = pair.split('=')
                unfurl.add_to_queue(
                    data_type='url.query.pair', key=key, value=value,
                    parent_id=node.node_id, incoming_edge_config=urlparse_edge)
            return



"""Module for encoding and decoding length delimited fields"""
import binascii
import copy
import sys
import six
import logging

from google.protobuf.internal import wire_format, encoder, decoder

from unfurl.lib import blackboxprotobuf
from unfurl.lib.blackboxprotobuf.lib.types import varint
from unfurl.lib.blackboxprotobuf.lib.exceptions import (
    EncoderException,
    DecoderException,
    TypedefException,
)


def encode_string(value):
    try:
        value = six.ensure_text(value)
    except TypeError as exc:
        six.raise_from(
            EncoderException("Error encoding string to message: %r" % value), exc
        )
    return encode_bytes(value)


def encode_bytes(value):
    """Encode varint length followed by the string.
    This should also work to encode incoming string values.
    """
    if isinstance(value, bytearray):
        value = bytes(value)
    try:
        value = six.ensure_binary(value)
    except TypeError as exc:
        six.raise_from(
            EncoderException("Error encoding bytes to message: %r" % value), exc
        )
    encoded_length = varint.encode_varint(len(value))
    return encoded_length + value


def decode_bytes(value, pos):
    """Decode varint for the length and then returns that number of bytes"""
    length, pos = varint.decode_varint(value, pos)
    end = pos + length
    try:
        return value[pos:end], end
    except IndexError as exc:
        six.raise_from(
            DecoderException(
                (
                    "Error decoding bytes. Decoded length %d is longer than bytes"
                    " available %d"
                )
                % (length, len(value) - pos)
            ),
            exc,
        )


def encode_bytes_hex(value):
    """Encode varint length followed by the string.
    Expects a string of hex characters
    """
    try:
        return encode_bytes(binascii.unhexlify(value))
    except (TypeError, binascii.Error) as exc:
        six.raise_from(
            EncoderException("Error encoding hex bytestring %s" % value), exc
        )


def decode_bytes_hex(buf, pos):
    """Decode varint for length and then returns that number of bytes.
    Outputs the bytes as a hex value
    """
    value, pos = decode_bytes(buf, pos)
    return binascii.hexlify(value), pos


def decode_string(value, pos):
    """Decode varint for length and then the bytes"""
    length, pos = varint.decode_varint(value, pos)
    end = pos + length
    try:
        # backslash escaping isn't reversible easily
        return value[pos:end].decode("utf-8"), end
    except (TypeError, UnicodeDecodeError) as exc:
        six.raise_from(
            DecoderException("Error decoding UTF-8 string %s" % value[pos:end]), exc
        )


def encode_message(data, config, typedef, path=None):
    """Encode a Python dictionary representing a protobuf message
    data - Python dictionary mapping field numbers to values
    typedef - Type information including field number, field name and field type
    This will throw an exception if an unkown value is used as a key
    """
    output = bytearray()
    if path is None:
        path = []

    for field_number, value in data.items():
        # Get the field number convert it as necessary
        alt_field_number = None

        string_types = str

        if isinstance(field_number, string_types):
            if "-" in field_number:
                field_number, alt_field_number = field_number.split("-")
            # TODO can refactor to cache the name to number mapping
            for number, info in typedef.items():
                if (
                    "name" in info
                    and info["name"] == field_number
                    and field_number != ""
                ):
                    field_number = number
                    break
        else:
            field_number = str(field_number)

        field_path = path[:]
        field_path.append(field_number)

        if field_number not in typedef:
            raise EncoderException(
                "Provided field name/number %s is not valid" % (field_number),
                field_path,
            )

        field_typedef = typedef[field_number]

        # Get encoder
        if "type" not in field_typedef:
            raise TypedefException(
                "Field %s does not have a defined type." % field_number, field_path
            )

        field_type = field_typedef["type"]

        field_encoder = None
        if alt_field_number is not None:
            if alt_field_number not in field_typedef["alt_typedefs"]:
                raise EncoderException(
                    "Provided alt field name/number %s is not valid for field_number %s"
                    % (alt_field_number, field_number),
                    field_path,
                )
            if isinstance(field_typedef["alt_typedefs"][alt_field_number], dict):
                innertypedef = field_typedef["alt_typedefs"][alt_field_number]
                field_encoder = lambda data: encode_lendelim_message(
                    data, config, innertypedef, path=field_path
                )

            else:
                # just let the field
                field_type = field_typedef["alt_typedefs"][alt_field_number]

        if field_encoder is None:
            if field_type == "message":
                innertypedef = None
                if "message_typedef" in field_typedef:
                    innertypedef = field_typedef["message_typedef"]
                elif "message_type_name" in field_typedef:
                    message_type_name = field_typedef["message_type_name"]
                    if message_type_name not in config.known_types:
                        raise TypedefException(
                            "Message type (%s) has not been defined"
                            % field_typedef["message_type_name"],
                            field_path,
                        )
                    innertypedef = config.known_types[message_type_name]
                else:
                    raise TypedefException(
                        "Could not find message typedef for %s" % field_number,
                        field_path,
                    )

                field_encoder = lambda data: encode_lendelim_message(
                    data, config, innertypedef, path=field_path
                )
            else:
                if field_type not in blackboxprotobuf.lib.types.ENCODERS:
                    raise TypedefException("Unknown type: %s" % field_type)
                field_encoder = blackboxprotobuf.lib.types.ENCODERS[field_type]
                if field_encoder is None:
                    raise TypedefException(
                        "Encoder not implemented for %s" % field_type, field_path
                    )

        # Encode the tag
        tag = encoder.TagBytes(
            int(field_number), blackboxprotobuf.lib.types.WIRETYPES[field_type]
        )

        try:
            # Handle repeated values
            if isinstance(value, list) and not field_type.startswith("packed_"):
                for repeated in value:
                    output += tag
                    output += field_encoder(repeated)
            else:
                output += tag
                output += field_encoder(value)
        except EncoderException as exc:
            exc.set_path(field_path)
            six.reraise(*sys.exc_info())

    return output


def decode_message(buf, config, typedef=None, pos=0, end=None, depth=0, path=None):
    """Decode a protobuf message with no length delimiter"""
    # TODO recalculate and re-add path for errors
    if end is None:
        end = len(buf)

    logging.debug("End: %d", end)
    if typedef is None:
        typedef = {}
    else:
        # Don't want to accidentally modify the original
        typedef = copy.deepcopy(typedef)

    if path is None:
        path = []

    output = {}

    grouped_fields, pos = _group_by_number(buf, pos, end, path)
    for (field_number, (wire_type, buffers)) in grouped_fields.items():
        # wire_type should already be validated by _group_by_number

        path = path[:] + [field_number]
        field_outputs = None
        field_typedef = typedef.get(field_number, {})
        field_key = _get_field_key(field_number, typedef, path)
        # Easy cases. Fixed size or bytes/string
        if (
            wire_type
            in [
                wire_format.WIRETYPE_FIXED32,
                wire_format.WIRETYPE_FIXED64,
                wire_format.WIRETYPE_VARINT,
            ]
            or ("type" in field_typedef and field_typedef["type"] != "message")
        ):

            if "type" not in field_typedef:
                field_typedef["type"] = config.get_default_type(wire_type)
            else:
                # have a type, but make sure it matches the wiretype
                if (
                    blackboxprotobuf.lib.types.WIRETYPES[field_typedef["type"]]
                    != wire_type
                ):
                    raise DecoderException(
                        "Type %s from typedef did not match wiretype %s for "
                        "field %s" % (field_typedef["type"], wire_type, field_key),
                        path=path,
                    )

            # we already have a type, just map the decoder
            if field_typedef["type"] not in blackboxprotobuf.lib.types.DECODERS:
                raise TypedefException(
                    "Got unkown type %s for field_number %s"
                    % (field_typedef["type"], field_number),
                    path=path,
                )

            decoder = blackboxprotobuf.lib.types.DECODERS[field_typedef["type"]]
            field_outputs = [decoder(buf, 0) for buf in buffers]

            # this shouldn't happen, but let's check just in case
            for buf, _pos in zip(buffers, [y for _, y in field_outputs]):
                assert len(buf) == _pos

            field_outputs = [value for (value, _) in field_outputs]
            if len(field_outputs) == 1:
                output[field_key] = field_outputs[0]
            else:
                output[field_key] = field_outputs

        elif wire_type == wire_format.WIRETYPE_LENGTH_DELIMITED:
            _try_decode_lendelim_fields(
                buffers, field_key, field_typedef, output, config
            )

        # Save the field typedef/type back to the typedef
        typedef[field_number] = field_typedef

    return output, typedef, pos


def _group_by_number(buf, pos, end, path):
    """Parse through the whole message and return buffers based on wire type.
    This forces us to parse the whole message at once, but I think we're
    doing that anyway.
    Returns a dictionary like:
        {
            "2": (<wiretype>, [<data>])
        }
    """

    output_map = {}
    while pos < end:
        # Read in a field
        try:
            if six.PY2:
                tag, pos = decoder._DecodeVarint(str(buf), pos)
            else:
                tag, pos = decoder._DecodeVarint(buf, pos)
        except (IndexError, decoder._DecodeError) as exc:
            six.raise_from(
                DecoderException(
                    "Error decoding length from buffer: %r..."
                    % (binascii.hexlify(buf[pos : pos + 8])),
                    path=path,
                ),
                exc,
            )

        field_number, wire_type = wire_format.UnpackTag(tag)

        # We want field numbers as strings everywhere
        field_number = str(field_number)

        path = path[:] + [field_number]

        if field_number in output_map and output_map[field_number][0] != wire_type:
            """This should never happen"""
            raise DecoderException(
                "Field %s has mistmatched wiretypes. Previous: %s Now: %s"
                % (field_number, output_map[field_number][0], wire_type),
                path=path,
            )

        length = None
        if wire_type == wire_format.WIRETYPE_VARINT:
            # We actually have to read in the whole varint to figure out it's size
            _, new_pos = varint.decode_varint(buf, pos)
            length = new_pos - pos
        elif wire_type == wire_format.WIRETYPE_FIXED32:
            length = 4
        elif wire_type == wire_format.WIRETYPE_FIXED64:
            length = 8
        elif wire_type == wire_format.WIRETYPE_LENGTH_DELIMITED:
            # Read the length from the start of the message
            # add on the length of the length tag as well
            bytes_length, new_pos = varint.decode_varint(buf, pos)
            length = bytes_length + (new_pos - pos)
        elif wire_type in [
            wire_format.WIRETYPE_START_GROUP,
            wire_format.WIRETYPE_END_GROUP,
        ]:
            raise DecoderException("GROUP wire types not supported", path=path)
        else:
            raise DecoderException("Got unkown wire type: %d" % wire_type, path=path)
        if pos + length > end:
            raise DecoderException(
                "Decoded length for field %s goes over end: %d > %d"
                % (field_number, pos + length, end),
                path=path,
            )

        field_buf = buf[pos : pos + length]

        if field_number in output_map:
            output_map[field_number][1].append(field_buf)
        else:
            output_map[field_number] = (wire_type, [field_buf])
        pos += length
    return output_map, pos


def _get_field_key(field_number, typedef, path):
    """If field_number has a name, then use that"""
    if not isinstance(field_number, (int, str)):
        # Should only get unpredictable inputs from encoding
        raise EncoderException("Field key in message must be a str or int", path=path)
    if isinstance(field_number, int):
        field_number = str(field_number)
    alt_field_number = None
    if "-" in field_number:
        field_number, alt_field_number = field_number.split("-")
        # TODO
        raise NotImplementedError(
            "Handling for _get_field_key not implemented for alt typedefs: %s"
            % field_number,
            path=path,
        )
    if field_number in typedef and "name" in typedef[field_number]:
        field_key = typedef[field_number]["name"]
    else:
        field_key = field_number
    # Return the new field_name + alt_field_number
    return field_key + ("" if alt_field_number is None else "-" + alt_field_number)


def _try_decode_lendelim_fields(
    buffers, field_key, field_typedef, message_output, config
):
    # This is where things get weird
    # To start, since we want to decode messages and not treat every
    # embedded message as bytes, we have to guess if it's a message or
    # not.
    # Unlike other types, we can't assume our message types are
    # consistent across the tree or even within the same message.
    # A type could be a bytes type that might decode to different
    # messages that don't have the same type definition. This is where
    # 'alt_typedefs' let us say, these are the different message types we've
    # seen for this one field.
    # In general, if something decodes as a message once, the rest should too
    # and we can enforce that across a single message, but not multiple
    # messages.
    # This is going to change the definition of "alt_typedefs" a bit from just
    # alternate message type definitions to also allowing downgrading to
    # 'bytes' or string with an 'alt_type' if it doesn't parse

    try:
        outputs_map = {}
        # grab all dictonary alt_typedefs
        all_typedefs = {
            # we don't want this to modify in-place if it fails
            key: copy.deepcopy(value)
            for key, value in field_typedef.get("alt_typedefs", {}).items()
            if isinstance(value, dict)
        }
        all_typedefs["1"] = copy.deepcopy(field_typedef.get("message_typedef", {}))

        for buf in buffers:
            output = None
            output_typedef = None
            output_typedef_num = None
            for alt_typedef_num, alt_typedef in sorted(
                all_typedefs.items(), key=lambda x: int(x[0])
            ):
                try:
                    output, output_typedef, _ = decode_lendelim_message(
                        buf, config, alt_typedef
                    )
                except:
                    continue
                output_typedef_num = alt_typedef_num
                break
            # try an anonymous type
            # let the error propogate up if we fail this
            if output is None:
                output, output_typedef, _ = decode_lendelim_message(buf, config, {})
                output_typedef_num = str(
                    max([int(i) for i in ["0"] + list(all_typedefs.keys())]) + 1
                )

            # save the output or typedef we found
            all_typedefs[output_typedef_num] = output_typedef
            output_list = outputs_map.get(output_typedef_num, [])
            output_list.append(output)
            outputs_map[output_typedef_num] = output_list
        # was able to decode everything as a message
        field_typedef["type"] = "message"
        field_typedef["message_typedef"] = all_typedefs["1"]
        if len(all_typedefs.keys()) > 1:
            del all_typedefs["1"]
            field_typedef.setdefault("alt_typedefs", {}).update(all_typedefs)
        # messages get set as "key-alt_number"
        for output_typedef_num, outputs in outputs_map.items():
            output_field_key = field_key
            if output_typedef_num != "1":
                output_field_key += "-" + output_typedef_num
            message_output[output_field_key] = (
                outputs if len(outputs) > 1 else outputs[0]
            )
        # success, return
        return
    except DecoderException as exc:
        # this should be pretty common, don't be noisy or throw an exception
        logging.debug(
            "Could not decode a buffer for field number %s as a message: %s",
            field_key,
            exc,
        )

    # Decoding as a message did not work, try strings and then bytes
    # The bytes decoding should never fail
    for target_type in ["string", config.default_binary_type]:
        try:
            outputs = []
            decoder = blackboxprotobuf.lib.types.DECODERS[target_type]
            for buf in buffers:
                output, _ = decoder(buf, 0)
                outputs.append(output)

            # all outputs worked, this is our type
            # check if there is a message type already in the typedef
            if "type" in field_typedef and "message" == field_typedef["type"]:
                # we already had a message type. save it as an alt_typedef

                # check if we already have this type as an alt_typedef
                output_typedef_nums = {
                    key: value
                    for key, value in field_typedef.setdefault(
                        "alt_typedefs", {}
                    ).items()
                    if value == target_type
                }.keys()
                output_typedef_num = None
                if len(output_typedef_nums) == 0:
                    # find the next largest alt typedef number to put this type as
                    output_typedef_num = str(
                        max([int(i) for i in ["0"] + all_typedefs.keys()]) + 1
                    )
                    field_typedef.setdefault("alt_typedefs", {})[
                        output_typedef_num
                    ] = target_type
                else:
                    # we already have an alt typedef with this number
                    output_typedef_num = output_typedef_nums[0]
                message_output[field_key + "-" + output_typedef_num] = (
                    outputs if len(outputs) > 1 else outputs[0]
                )
            else:
                field_typedef["type"] = target_type
                message_output[field_key] = outputs if len(outputs) > 1 else outputs[0]
            return
        except DecoderException:
            continue


def encode_lendelim_message(data, config, typedef, path=None):
    """Encode the length before the message"""
    message_out = encode_message(data, config, typedef, path=path)
    length = varint.encode_varint(len(message_out))
    logging.debug("Message length encoded: %d", len(length) + len(message_out))
    return length + message_out


def decode_lendelim_message(buf, config, typedef=None, pos=0, depth=0, path=None):
    """Read in the length and use it as the end"""
    length, pos = varint.decode_varint(buf, pos)
    ret = decode_message(
        buf, config, typedef, pos, pos + length, depth=depth, path=path
    )
    return ret


def generate_packed_encoder(wrapped_encoder):
    """Generate an encoder for a packed type from the base type encoder"""

    def length_wrapper(values):
        """Encode repeat values and prefix with the length"""
        output = bytearray()
        for value in values:
            output += wrapped_encoder(value)
        length = varint.encode_varint(len(output))
        return length + output

    return length_wrapper


def generate_packed_decoder(wrapped_decoder):
    """Generate an decoder for a packer type from a base type decoder"""

    def length_wrapper(buf, pos):
        """Decode repeat values prefixed with the length"""
        length, pos = varint.decode_varint(buf, pos)
        end = pos + length
        output = []
        while pos < end:
            value, pos = wrapped_decoder(buf, pos)
            output.append(value)
        if pos > end:
            raise DecoderException(
                (
                    "Error decoding packed field. Packed length larger than"
                    " buffer: decoded = %d, left = %d"
                )
                % (length, len(buf) - pos)
            )
        return output, pos

    return length_wrapper



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
from unfurl.lib import blackboxprotobuf
from unfurl import utils

proto_edge = {
    'color': {
        'color': '#2D9A58'
    },
    'title': 'Protobuf Parsing Functions',
    'label': 'proto'
}

hex_proto_edge = {
    'color': {
        'color': '#2D9A58'
    },
    'title': 'Parsed from hex bytes, then as a protobuf',
    'label': 'hex+proto'
}

b64_proto_edge = {
    'color': {
        'color': '#2D9A58'
    },
    'title': 'Parsed from base64, then as a protobuf',
    'label': 'b64+proto'
}

wire_types = {
    'int': '<b>varint (0)</b>. <br><br>It is displayed here as an int, but it could orginially have '
           'been <br>a int32, int64, uint32, uint64, sint32, sint64, bool, or enum.',
    'fixed64': '<b>64-bit (1)</b>. <br><br>It is displayed here as a fixed64, but it could originally '
               'have been <br>a fixed64, sfixed64, or double.',
    'bytes': '<b>length-delimited (2)</b>. <br><br>It is displayed here as bytes, but it could originally '
             'have been <br>a string, bytes, embedded messages, or packed repeated fields.',
    'string': '<b>length-delimited (2)</b>. <br><br>It is displayed here as a string, but it could originally '
             'have been <br>a string, bytes, embedded messages, or packed repeated fields.',
    'fixed32': '<b>32-bit (5)</b>. <br><br>It is displayed here as an int, but it could originally '
               'have been <br>a fixed32, sfixed32, or float.'
}


def run(unfurl, node):

    def parse_protobuf_into_nodes(pb_value_dict, pb_types, edge_type=None):
        assert isinstance(pb_value_dict, dict), \
            f'"parse_protobuf_into_nodes" expects a dict, but got {type(pb_value_dict)} as input'

        if len(pb_value_dict) > 0:
            for field_number, field_value in pb_value_dict.items():
                if isinstance(field_value, (str, int, float, bytes, bytearray)):
                    unfurl.add_to_queue(
                        data_type='proto', key=field_number, value=str(field_value),
                        hover=f'Field number <b>{field_number}</b> has a wire '
                              f'type of {wire_types[pb_types[field_number]["type"]]}',
                        parent_id=node.node_id, incoming_edge_config=edge_type)
                elif isinstance(field_value, dict):
                    unfurl.add_to_queue(
                        data_type='proto.dict', key=field_number,
                        value={'field_values': field_value, 'field_types': pb_types[field_number]["message_typedef"]},
                        label=f'{field_number}: {field_value}',
                        hover=f'Field number <b>{field_number}</b> '
                              f'is a nested message; it will be parsed further into more nodes',
                        parent_id=node.node_id, incoming_edge_config=edge_type)
                elif isinstance(field_value, list):
                    nested_types = pb_types[field_number]
                    if pb_types[field_number].get("message_typedef"):
                        nested_types = pb_types[field_number]["message_typedef"]

                    unfurl.add_to_queue(
                        data_type='proto.list', key=field_number,
                        value={'field_values': field_value, 'field_types': nested_types},
                        label=f'{field_number}: {field_value}',
                        hover=f'Field number <b>{field_number}</b> '
                              f'is a nested message; it will be parsed further into more nodes',
                        parent_id=node.node_id, incoming_edge_config=edge_type)

    if node.data_type == 'proto.dict':
        parse_protobuf_into_nodes(node.value.get('field_values'), node.value.get('field_types'), proto_edge)
        return

    if node.data_type == 'proto.list':
        for field_value in node.value['field_values']:
            field_types = node.value.get('field_types')
            if not isinstance(field_value, dict):
                field_value = {node.key: field_value}
                field_types = {node.key: field_types}
            parse_protobuf_into_nodes(field_value, field_types, proto_edge)
        return

    if node.data_type == 'bytes':
        try:
            protobuf_values, protobuf_values_types = blackboxprotobuf.decode_message(node.value)
            parse_protobuf_into_nodes(protobuf_values, protobuf_values_types, proto_edge)
            return

        # This will often fail for a wide array of reasons when it tries to parse a non-pb as a pb
        except Exception:
            pass

    if not isinstance(node.value, str):
        return False

    if len(node.value) % 4 == 1:
        # A valid b64 string will not be this length
        return False

    if node.data_type.startswith(('uuid', 'hash')):
        return False

    urlsafe_b64_m = utils.urlsafe_b64_re.fullmatch(node.value)
    standard_b64_m = utils.standard_b64_re.fullmatch(node.value)
    hex_m = utils.hex_re.fullmatch(node.value)
    long_int_m = utils.long_int_re.fullmatch(node.value)
    all_digits_m = utils.digits_re.fullmatch(node.value)
    all_letters_m = utils.digits_re.fullmatch(node.value)

    if hex_m and not (all_digits_m or all_letters_m):
        decoded = bytes.fromhex(node.value)
        try:
            protobuf_values, protobuf_values_types = blackboxprotobuf.decode_message(decoded)
            parse_protobuf_into_nodes(protobuf_values, protobuf_values_types, hex_proto_edge)
            return

        # This will often fail for a wide array of reasons when it tries to parse a non-pb as a pb
        except Exception:
            return

    elif urlsafe_b64_m and not (all_digits_m or all_letters_m):
        try:
            decoded = base64.urlsafe_b64decode(unfurl.add_b64_padding(node.value))
            protobuf_values, protobuf_values_types = blackboxprotobuf.decode_message(decoded)
            parse_protobuf_into_nodes(protobuf_values, protobuf_values_types, b64_proto_edge)
            return

        # This will often fail for a wide array of reasons when it tries to parse a non-pb as a pb
        except Exception:
            return

    elif standard_b64_m and not (all_digits_m or all_letters_m):
        try:
            decoded = base64.b64decode(unfurl.add_b64_padding(node.value))
            protobuf_values, protobuf_values_types = blackboxprotobuf.decode_message(decoded)
            parse_protobuf_into_nodes(protobuf_values, protobuf_values_types, b64_proto_edge)
            return

        # This will often fail for a wide array of reasons when it tries to parse a non-pb as a pb
        except Exception:
            return



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

mailto_edge = {
    'color': {'color': 'lightblue'},
    'title': 'Mailto Parsing Functions',
    'label': '📧',
}


def run(unfurl, node):
    if node.data_type == 'url' and node.value.startswith('mailto:'):

        raw = node.value.split(':', 1)

        if len(raw[1]) == 0:
            return

        nodes = []

        if '?' in raw[1]:
            split_qs = raw[1].split('?', 1)

            if len(split_qs[0]) > 0:
                nodes.append({'k': 'to', 'v': split_qs[0]})

            # rest of the string for node value, parse as 'url.query;'
            for pair in split_qs[1].replace('&amp;', '&').split('&'):
                r = pair.split('=', 1)
                nodes.append({'k': r[0], 'v': r[1]})
        else:
            nodes.append({'k': 'to', 'v': raw[1]})

        for n in nodes:
            k = n['k']
            v = n['v']

            if k in ['to', 'cc', 'bcc']:
                if ',' in v:
                    data_type = 'mailto.email.multiple'
                else:
                    data_type = 'email'
                unfurl_val = v
            else:
                data_type = 'url.query'
                unfurl_val = f'{k}={v}'

            unfurl.add_to_queue(
                data_type=data_type, key=k, value=unfurl_val, label=f'{k}: {v}',
                hover='Mailto link is a type of URL that opens the default mail client for sending an email, per '
                      '<a href="https://tools.ietf.org/html/rfc2368" target="_blank">RFC2368</a>',
                parent_id=node.node_id, incoming_edge_config=mailto_edge)

    if node.data_type == 'mailto.email.multiple':
        for raw_email in node.value.split(','):
            email = raw_email.replace('%20', ' ').strip()
            unfurl.add_to_queue(
                data_type='email', key=None, value=email, label=f'{node.key}: {email}',
                hover='Mailto link is a type of URL that opens the default mail client for sending an email, '
                      'per <a href="https://tools.ietf.org/html/rfc2368" target="_blank">RFC2368</a>',
                parent_id=node.node_id, incoming_edge_config=mailto_edge)



"""Classes for encoding and decoding varint types"""
import binascii
import struct

from google.protobuf.internal import wire_format, encoder, decoder
import six

from unfurl.lib.blackboxprotobuf.lib.exceptions import EncoderException, DecoderException


def _gen_append_bytearray(arr):
    def _append_bytearray(x):
        if isinstance(x, (str, int)):
            arr.append(x)
        elif isinstance(x, bytes):
            arr.extend(x)
        else:
            raise EncoderException("Unknown type returned by protobuf library")

    return _append_bytearray


def encode_uvarint(value):
    """Encode a long or int into a bytearray."""
    output = bytearray()
    if value < 0:
        raise EncoderException(
            "Error encoding %d as uvarint. Value must be positive" % value
        )
    try:
        encoder._EncodeVarint(_gen_append_bytearray(output), value)
    except (struct.error, ValueError) as exc:
        six.raise_from(EncoderException("Error encoding %d as uvarint." % value), exc)

    return output


def decode_uvarint(buf, pos):
    """Decode bytearray into a long."""
    # Convert buffer to string
    if six.PY2:
        buf = str(buf)
    try:
        value, pos = decoder._DecodeVarint(buf, pos)
    except (TypeError, IndexError, decoder._DecodeError) as exc:
        six.raise_from(
            DecoderException(
                "Error decoding uvarint from %s..."
                % binascii.hexlify(buf[pos : pos + 8])
            ),
            exc,
        )
    return (value, pos)


def encode_varint(value):
    """Encode a long or int into a bytearray."""
    output = bytearray()
    if value > (2 ** 63) or value < -(2 ** 63):
        raise EncoderException("Value %d above maximum varint size" % value)
    try:
        encoder._EncodeSignedVarint(_gen_append_bytearray(output), value)
    except (struct.error, ValueError) as exc:
        six.raise_from(
            EncoderException("Error encoding %d as signed varint." % value), exc
        )
    return output


def decode_varint(buf, pos):
    """Decode bytearray into a long."""
    # Convert buffer to string
    if six.PY2:
        buf = str(buf)
    try:
        value, pos = decoder._DecodeSignedVarint(buf, pos)
    except (TypeError, IndexError, decoder._DecodeError) as exc:
        six.raise_from(
            DecoderException(
                "Error decoding varint from %s..."
                % binascii.hexlify(buf[pos : pos + 8])
            ),
            exc,
        )
    return (value, pos)


def encode_svarint(value):
    """Zigzag encode the potentially signed value prior to encoding"""
    # zigzag encode value
    return encode_uvarint(wire_format.ZigZagEncode(value))


def decode_svarint(buf, pos):
    """Decode bytearray into a long."""
    output, pos = decode_uvarint(buf, pos)
    # zigzag encode value
    return wire_format.ZigZagDecode(output), pos
